"""Tests for MCP server tools (without actual MCP connection)."""

import pytest


def test_tool_validate_workflow_valid():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    yaml_content = """
version: "1"
start_at: translate
end_at:
  - translate
nodes:
  - id: translate
    handler: llm_handler
edges: []
"""
    result = _tool_validate_workflow(yaml_content)
    assert result["is_valid"] is True
    assert result["issues"] == []


def test_tool_validate_workflow_invalid_node_ref():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    yaml_content = """
version: "1"
start_at: translat
end_at:
  - translate
nodes:
  - id: translate
    handler: llm_handler
edges: []
"""
    result = _tool_validate_workflow(yaml_content)
    assert result["is_valid"] is False
    assert len(result["issues"]) > 0


def test_tool_validate_workflow_yaml_error():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    result = _tool_validate_workflow("not: valid: yaml: [")
    assert result["is_valid"] is False


def test_tool_explain_workflow():
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    yaml_content = """
version: "1"
start_at: translate
end_at:
  - translate
nodes:
  - id: translate
    handler: llm
    params:
      output_key: translation
edges: []
"""
    result = _tool_explain_workflow(yaml_content)
    assert result["entry_point"] == "translate"
    assert "llm" in result["required_handlers"]
    assert "translation" in result["variable_flow"]["translate"]["outputs"]


def test_tool_list_templates():
    from yagra.adapters.inbound.mcp_server import _tool_list_templates

    result = _tool_list_templates()
    assert "templates" in result
    assert isinstance(result["templates"], list)
    # Each template entry must contain name, description, and use_case keys
    for tmpl in result["templates"]:
        assert "name" in tmpl, f"missing 'name' in template entry: {tmpl}"
        assert "description" in tmpl, f"missing 'description' in template entry: {tmpl}"
        assert "use_case" in tmpl, f"missing 'use_case' in template entry: {tmpl}"


def test_tool_list_handlers():
    from yagra.adapters.inbound.mcp_server import _tool_list_handlers

    result = _tool_list_handlers()
    assert "handlers" in result
    names = [h["name"] for h in result["handlers"]]
    assert "llm" in names
    assert "structured_llm" in names
    assert "streaming_llm" in names


def test_create_mcp_server_requires_mcp_package():
    """Confirms that an ImportError is raised when the mcp package is not installed."""
    # This test confirms that a server object is returned if mcp is installed.
    # If mcp is not installed, an ImportError is raised (which is also correct behavior).
    try:
        from yagra.adapters.inbound.mcp_server import create_mcp_server

        server = create_mcp_server()
        assert server is not None
    except ImportError:
        # mcp not installed is a normal case
        pass


# --- _tool_validate_workflow: payload is not a dict (line 148-154) ---


def test_tool_validate_workflow_non_mapping_payload():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    result = _tool_validate_workflow("- item1\n- item2\n")
    assert result["is_valid"] is False
    issues = result["issues"]
    assert len(issues) > 0
    assert any(i["code"] == "schema_error" for i in issues)


def test_tool_validate_workflow_null_payload():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    result = _tool_validate_workflow("null\n")
    assert result["is_valid"] is False


# --- _tool_explain_workflow: ImportError case (line 183-184) ---


def test_tool_explain_workflow_import_error(monkeypatch):
    import sys

    monkeypatch.setitem(
        sys.modules,
        "yagra.application.use_cases.workflow_explainer",
        None,
    )

    import importlib

    import yagra.adapters.inbound.mcp_server as mcp_module

    importlib.reload(mcp_module)

    yaml_content = """
version: "1"
start_at: translate
end_at:
  - translate
nodes:
  - id: translate
    handler: llm_handler
edges: []
"""
    result = mcp_module._tool_explain_workflow(yaml_content)
    assert "error" in result
    assert "explain_workflow" in result["error"] or "not yet available" in result["error"]

    importlib.reload(mcp_module)


# --- _tool_explain_workflow: YAML parse error (line 188-189) ---


def test_tool_explain_workflow_yaml_error():
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    result = _tool_explain_workflow("not: valid: yaml: [")
    assert "error" in result
    assert "YAML" in result["error"]


# --- _tool_explain_workflow: payload is not a dict (line 192) ---


def test_tool_explain_workflow_non_mapping_payload():
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    result = _tool_explain_workflow("- item1\n- item2\n")
    assert result == {"error": "workflow must be a mapping"}


def test_tool_explain_workflow_null_payload():
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    result = _tool_explain_workflow("null\n")
    assert result == {"error": "workflow must be a mapping"}


# --- _tool_explain_workflow: validation failure (line 200) ---


def test_tool_explain_workflow_validation_failure():
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    yaml_content = """
version: "1"
start_at: nonexistent_node
end_at:
  - translate
nodes:
  - id: translate
    handler: llm_handler
edges: []
"""
    result = _tool_explain_workflow(yaml_content)
    assert "error" in result
    assert "validation" in result["error"] or "issues" in result


# --- create_mcp_server call_tool (only if mcp is installed) ---


def test_create_mcp_server_call_tool_if_mcp_installed():
    import asyncio

    try:
        from yagra.adapters.inbound.mcp_server import create_mcp_server
    except ImportError:
        return

    try:
        server = create_mcp_server()
    except ImportError:
        return

    call_tool_handler = None
    for handler_name, handler_fn in server.request_handlers.items():
        if "CallTool" in str(handler_name):
            call_tool_handler = handler_fn
            break

    if call_tool_handler is None:
        return

    async def run_validate():
        from mcp.types import CallToolRequest, CallToolRequestParams

        request = CallToolRequest(
            method="tools/call",
            params=CallToolRequestParams(
                name="validate_workflow",
                arguments={
                    "yaml_content": (
                        "version: '1'\nstart_at: n\nend_at: [n]\nnodes:\n  - id: n\n    handler: h\nedges: []\n"
                    )
                },
            ),
        )
        result = await call_tool_handler(request)
        return result

    try:
        result = asyncio.run(run_validate())
        assert result is not None
    except Exception:
        pass


def test_create_mcp_server_call_tool_unknown_tool():
    import asyncio

    try:
        from yagra.adapters.inbound.mcp_server import create_mcp_server

        server = create_mcp_server()
    except ImportError:
        return

    call_tool_handler = None
    for handler_name, handler_fn in server.request_handlers.items():
        if "CallTool" in str(handler_name):
            call_tool_handler = handler_fn
            break

    if call_tool_handler is None:
        return

    async def run_unknown():
        from mcp.types import CallToolRequest, CallToolRequestParams

        request = CallToolRequest(
            method="tools/call",
            params=CallToolRequestParams(name="unknown_tool", arguments={}),
        )
        result = await call_tool_handler(request)
        return result

    try:
        import json

        result = asyncio.run(run_unknown())

        assert result is not None
        contents = result.root.content if hasattr(result, "root") else []
        if contents:
            data = json.loads(contents[0].text)
            assert "error" in data
    except Exception:
        pass


# ---------------------------------------------------------------------------
# _resolve_mcp_workflow_path
# ---------------------------------------------------------------------------


def test_resolve_mcp_workflow_path_none():
    from pathlib import Path

    from yagra.adapters.inbound.mcp_server import _resolve_mcp_workflow_path

    assert _resolve_mcp_workflow_path(None) == Path("<mcp>")


def test_resolve_mcp_workflow_path_with_dir():
    from pathlib import Path

    from yagra.adapters.inbound.mcp_server import _resolve_mcp_workflow_path

    result = _resolve_mcp_workflow_path("/tmp/mydir")
    assert result == Path("/tmp/mydir/_mcp_workflow.yaml")
    assert result.parent == Path("/tmp/mydir")


# ---------------------------------------------------------------------------
# _tool_validate_workflow with workflow_path
# ---------------------------------------------------------------------------

_VALID_WORKFLOW_YAML = """\
version: "1"
start_at: node1
end_at:
  - node1
nodes:
  - id: node1
    handler: my_handler
edges: []
"""


def test_tool_validate_workflow_with_workflow_path_valid(tmp_path):
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML)

    result = _tool_validate_workflow(workflow_path=str(wf))
    assert result["is_valid"] is True
    assert result["issues"] == []


def test_tool_validate_workflow_with_workflow_path_not_found():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    result = _tool_validate_workflow(workflow_path="/nonexistent/path/workflow.yaml")
    assert result["is_valid"] is False
    assert any(i["code"] == "schema_error" for i in result["issues"])
    assert any("not found" in i["message"] for i in result["issues"])


def test_tool_validate_workflow_with_workflow_path_resolves_prompt_ref(tmp_path):
    """prompt_ref relative paths are resolved from the workflow file's directory."""
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    prompts_dir = tmp_path / "prompts"
    prompts_dir.mkdir()
    prompt_file = prompts_dir / "chat.yaml"
    prompt_file.write_text("system: You are a helpful assistant.\n")

    wf = tmp_path / "workflow.yaml"
    wf.write_text(
        """\
version: "1"
start_at: chat
end_at:
  - chat
nodes:
  - id: chat
    handler: llm
    params:
      prompt_ref: "prompts/chat.yaml#system"
      model:
        provider: openai
        name: gpt-4o-mini
edges: []
"""
    )

    result = _tool_validate_workflow(workflow_path=str(wf))
    assert result["is_valid"] is True, result


def test_tool_validate_workflow_neither_provided():
    """Returns error when neither yaml_content nor workflow_path is provided."""
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    result = _tool_validate_workflow()
    assert result["is_valid"] is False
    assert any("either yaml_content or workflow_path" in i["message"] for i in result["issues"])


# ---------------------------------------------------------------------------
# _tool_explain_workflow with workflow_path
# ---------------------------------------------------------------------------


def test_tool_explain_workflow_with_workflow_path_valid(tmp_path):
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML)

    result = _tool_explain_workflow(workflow_path=str(wf))
    assert result["entry_point"] == "node1"
    assert "my_handler" in result["required_handlers"]


def test_tool_explain_workflow_with_workflow_path_not_found():
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    result = _tool_explain_workflow(workflow_path="/nonexistent/path/workflow.yaml")
    assert "error" in result
    assert "not found" in result["error"]


def test_tool_explain_workflow_neither_provided():
    """Returns error when neither yaml_content nor workflow_path is provided."""
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    result = _tool_explain_workflow()
    assert "error" in result
    assert "either yaml_content or workflow_path" in result["error"]


# ---------------------------------------------------------------------------
# base_dir parameter on yaml_content tools
# ---------------------------------------------------------------------------


def test_tool_validate_workflow_with_base_dir_resolves_prompt_ref(tmp_path):
    """base_dir enables relative prompt_ref resolution when passing yaml_content."""
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    prompts_dir = tmp_path / "prompts"
    prompts_dir.mkdir()
    (prompts_dir / "chat.yaml").write_text("system: You are a helpful assistant.\n")

    yaml_content = """\
version: "1"
start_at: chat
end_at:
  - chat
nodes:
  - id: chat
    handler: llm
    params:
      prompt_ref: "prompts/chat.yaml#system"
      model:
        provider: openai
        name: gpt-4o-mini
edges: []
"""
    result = _tool_validate_workflow(yaml_content, base_dir=str(tmp_path))
    assert result["is_valid"] is True, result


# ---------------------------------------------------------------------------
# _tool_get_traces
# ---------------------------------------------------------------------------

_SAMPLE_TRACE = {
    "schema_version": "1.0",
    "run_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
    "workflow_name": "translate",
    "workflow_version": "1.0",
    "started_at": "2026-02-20T10:00:00+00:00",
    "ended_at": "2026-02-20T10:00:01+00:00",
    "status": "success",
    "nodes": [
        {
            "node_id": "translate",
            "handler": "llm",
            "status": "success",
            "started_at": "2026-02-20T10:00:00+00:00",
            "ended_at": "2026-02-20T10:00:01+00:00",
            "duration_ms": 950.0,
            "input_snapshot": {"input": "hello"},
            "output_snapshot": {"translation": "こんにちは"},
            "llm_call": {
                "model": "openai/gpt-4o-mini",
                "provider": "openai",
                "prompt_tokens": 50,
                "completion_tokens": 20,
                "total_tokens": 70,
                "estimated_cost_usd": 0.00015,
            },
            "error": None,
        }
    ],
    "summary": {
        "total_nodes_executed": 1,
        "succeeded_nodes": 1,
        "failed_nodes": 0,
        "total_duration_ms": 1000.0,
        "total_prompt_tokens": 50,
        "total_completion_tokens": 20,
        "total_tokens": 70,
        "total_estimated_cost_usd": 0.00015,
        "node_order": ["translate"],
    },
    "metadata": {
        "yagra_version": "0.6.11",
        "python_version": "3.12.0",
        "platform": "Darwin",
    },
}


def _write_sample_trace(tmp_path, workflow_name="translate", trace_data=None):
    """Helper to write a sample trace JSON file in the LocalTraceSink directory layout."""
    import json

    data = trace_data if trace_data is not None else dict(_SAMPLE_TRACE)
    data["workflow_name"] = workflow_name
    workflow_dir = tmp_path / workflow_name
    workflow_dir.mkdir(parents=True, exist_ok=True)
    run_id = data.get("run_id", "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")
    run_id_short = run_id.replace("-", "")[:8]
    filename = f"{workflow_name}_20260220T100000_{run_id_short}.json"
    filepath = workflow_dir / filename
    filepath.write_text(json.dumps(data, ensure_ascii=False, indent=2), encoding="utf-8")
    return filepath


def test_get_traces_tool_listed():
    """Verify get_traces appears in list_tools."""
    try:
        from yagra.adapters.inbound.mcp_server import create_mcp_server
    except ImportError:
        return

    try:
        server = create_mcp_server()
    except ImportError:
        return

    import asyncio

    list_tools_handler = None
    for handler_name, handler_fn in server.request_handlers.items():
        if "ListTools" in str(handler_name):
            list_tools_handler = handler_fn
            break

    if list_tools_handler is None:
        return

    async def run():
        from mcp.types import ListToolsRequest

        request = ListToolsRequest(method="tools/list")
        result = await list_tools_handler(request)
        return result

    try:
        result = asyncio.run(run())
        tools = result.root.tools if hasattr(result, "root") else []
        tool_names = [t.name for t in tools]
        assert "get_traces" in tool_names
    except Exception:
        # If MCP internals differ, fall back to testing the implementation directly
        pass


def test_analyze_traces_tool_listed():
    """Verify analyze_traces appears in list_tools."""
    try:
        from yagra.adapters.inbound.mcp_server import create_mcp_server
    except ImportError:
        return

    try:
        server = create_mcp_server()
    except ImportError:
        return

    import asyncio

    list_tools_handler = None
    for handler_name, handler_fn in server.request_handlers.items():
        if "ListTools" in str(handler_name):
            list_tools_handler = handler_fn
            break

    if list_tools_handler is None:
        return

    async def run():
        from mcp.types import ListToolsRequest

        request = ListToolsRequest(method="tools/list")
        result = await list_tools_handler(request)
        return result

    try:
        result = asyncio.run(run())
        tools = result.root.tools if hasattr(result, "root") else []
        tool_names = [t.name for t in tools]
        assert "analyze_traces" in tool_names
    except Exception:
        pass


def test_get_traces_returns_traces(tmp_path):
    """Test get_traces with sample trace files in tmp_path."""
    from yagra.adapters.inbound.mcp_server import _tool_get_traces

    _write_sample_trace(tmp_path, workflow_name="translate")

    result = _tool_get_traces(trace_dir=str(tmp_path), workflow_name="translate", limit=20)
    assert "traces" in result
    assert "count" in result
    assert result["count"] == 1
    assert len(result["traces"]) == 1
    assert result["traces"][0]["workflow_name"] == "translate"


def test_get_traces_empty_dir(tmp_path):
    """Test get_traces with no trace files."""
    from yagra.adapters.inbound.mcp_server import _tool_get_traces

    result = _tool_get_traces(trace_dir=str(tmp_path), workflow_name=None, limit=20)
    assert "traces" in result
    assert result["count"] == 0
    assert result["traces"] == []


def test_get_traces_nonexistent_dir():
    """Test get_traces with a directory that does not exist."""
    from yagra.adapters.inbound.mcp_server import _tool_get_traces

    result = _tool_get_traces(trace_dir="/nonexistent/path/traces", workflow_name=None, limit=20)
    assert "traces" in result
    assert result["count"] == 0


def test_get_traces_filters_by_workflow_name(tmp_path):
    """Test that workflow_name filter selects the correct subdirectory."""
    from yagra.adapters.inbound.mcp_server import _tool_get_traces

    _write_sample_trace(tmp_path, workflow_name="translate")

    other_trace = dict(_SAMPLE_TRACE)
    other_trace["workflow_name"] = "summarize"
    other_trace["run_id"] = "11111111-2222-3333-4444-555555555555"
    _write_sample_trace(tmp_path, workflow_name="summarize", trace_data=other_trace)

    result = _tool_get_traces(trace_dir=str(tmp_path), workflow_name="translate", limit=20)
    assert result["count"] == 1
    assert result["traces"][0]["workflow_name"] == "translate"


def test_get_traces_respects_limit(tmp_path):
    """Test that the limit parameter caps the number of returned traces."""
    import json
    import uuid

    from yagra.adapters.inbound.mcp_server import _tool_get_traces

    for i in range(5):
        trace_data = dict(_SAMPLE_TRACE)
        trace_data["run_id"] = str(uuid.uuid4())
        trace_data["started_at"] = f"2026-02-20T10:00:{i:02d}+00:00"
        workflow_dir = tmp_path / "translate"
        workflow_dir.mkdir(parents=True, exist_ok=True)
        run_id = str(trace_data["run_id"])
        rid = run_id.replace("-", "")[:8]
        fp = workflow_dir / f"translate_20260220T1000{i:02d}_{rid}.json"
        fp.write_text(json.dumps(trace_data, ensure_ascii=False), encoding="utf-8")

    result = _tool_get_traces(trace_dir=str(tmp_path), workflow_name="translate", limit=3)
    assert result["count"] == 3


def test_analyze_traces_returns_summary(tmp_path):
    """Test aggregation with sample traces."""
    from yagra.adapters.inbound.mcp_server import _tool_analyze_traces

    _write_sample_trace(tmp_path, workflow_name="translate")

    result = _tool_analyze_traces(trace_dir=str(tmp_path), workflow_name="translate")
    assert "error" not in result
    assert result["total_runs"] == 1
    assert result["workflow_name"] == "translate"
    assert result["successful_runs"] == 1
    assert result["total_tokens"] == 70
    # Check per-node stats
    assert len(result["node_stats"]) == 1
    node = result["node_stats"][0]
    assert node["node_id"] == "translate"
    assert node["execution_count"] == 1
    assert node["success_count"] == 1


def test_analyze_traces_no_traces(tmp_path):
    """Test error response when no traces found."""
    from yagra.adapters.inbound.mcp_server import _tool_analyze_traces

    result = _tool_analyze_traces(trace_dir=str(tmp_path), workflow_name="nonexistent")
    assert "error" in result
    assert result["error"] == "No traces found"
    assert result["trace_dir"] == str(tmp_path)
    assert result["workflow_name"] == "nonexistent"


# ---------------------------------------------------------------------------
# _tool_propose_update
# ---------------------------------------------------------------------------

_VALID_WORKFLOW_YAML_PROPOSE = """\
version: "1"
start_at: node1
end_at:
  - node1
nodes:
  - id: node1
    handler: my_handler
edges: []
"""

_UPDATED_WORKFLOW_YAML_PROPOSE = """\
version: "1"
start_at: node1
end_at:
  - node1
nodes:
  - id: node1
    handler: my_handler
  - id: node2
    handler: other_handler
edges: []
"""


def test_tool_propose_update_valid(tmp_path):
    """Valid candidate YAML returns is_valid=True and a non-empty diff."""
    from yagra.adapters.inbound.mcp_server import _tool_propose_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    result = _tool_propose_update(
        workflow_path=str(wf),
        candidate_yaml=_UPDATED_WORKFLOW_YAML_PROPOSE,
        reason="add node2",
    )

    assert "error" not in result
    assert result["is_valid"] is True
    assert result["issues"] == []
    assert result["diff"] != ""
    assert result["reason"] == "add node2"
    assert result["current_yaml"] == _VALID_WORKFLOW_YAML_PROPOSE
    assert result["candidate_yaml"] == _UPDATED_WORKFLOW_YAML_PROPOSE


def test_tool_propose_update_invalid_yaml(tmp_path):
    """Malformed candidate_yaml returns an error key."""
    from yagra.adapters.inbound.mcp_server import _tool_propose_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    result = _tool_propose_update(
        workflow_path=str(wf),
        candidate_yaml="not: valid: yaml: [",
    )

    assert "error" in result


def test_tool_propose_update_no_change(tmp_path):
    """Same content as current file results in an empty diff string."""
    from yagra.adapters.inbound.mcp_server import _tool_propose_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    result = _tool_propose_update(
        workflow_path=str(wf),
        candidate_yaml=_VALID_WORKFLOW_YAML_PROPOSE,
    )

    assert "error" not in result
    assert result["diff"] == ""


def test_tool_propose_update_file_not_found():
    """Non-existent workflow_path returns an error key."""
    from yagra.adapters.inbound.mcp_server import _tool_propose_update

    result = _tool_propose_update(
        workflow_path="/nonexistent/path/workflow.yaml",
        candidate_yaml=_VALID_WORKFLOW_YAML_PROPOSE,
    )

    assert "error" in result


# ---------------------------------------------------------------------------
# _tool_apply_update
# ---------------------------------------------------------------------------


def test_tool_apply_update_success(tmp_path, monkeypatch):
    """Valid YAML is applied and success=True with backup_id is returned."""
    from yagra.adapters.inbound import mcp_server
    from yagra.application.use_cases.workflow_persistence import WorkflowSaveResult

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    mock_result = WorkflowSaveResult(saved_revision="abc123", backup_id="bk-001")
    monkeypatch.setattr(mcp_server, "save_workflow_with_backup", lambda *a, **k: mock_result)

    result = mcp_server._tool_apply_update(
        workflow_path=str(wf),
        candidate_yaml=_UPDATED_WORKFLOW_YAML_PROPOSE,
    )

    assert result.get("success") is True
    assert result["backup_id"] == "bk-001"
    assert result["saved_revision"] == "abc123"


def test_tool_apply_update_invalid_yaml(tmp_path):
    """Malformed candidate_yaml returns an error key."""
    from yagra.adapters.inbound.mcp_server import _tool_apply_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    result = _tool_apply_update(
        workflow_path=str(wf),
        candidate_yaml="not: valid: yaml: [",
    )

    assert "error" in result


def test_tool_apply_update_revision_conflict(tmp_path, monkeypatch):
    """WorkflowRevisionConflictError is mapped to error='revision_conflict'."""
    from yagra.adapters.inbound import mcp_server
    from yagra.application.use_cases.workflow_persistence import WorkflowRevisionConflictError

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    def raise_conflict(*args: object, **kwargs: object) -> None:
        raise WorkflowRevisionConflictError(
            expected_revision="expected_rev", actual_revision="actual_rev"
        )

    monkeypatch.setattr(mcp_server, "save_workflow_with_backup", raise_conflict)

    result = mcp_server._tool_apply_update(
        workflow_path=str(wf),
        candidate_yaml=_UPDATED_WORKFLOW_YAML_PROPOSE,
        base_revision="expected_rev",
    )

    assert result.get("error") == "revision_conflict"
    assert "expected" in result
    assert "actual" in result


def test_tool_apply_update_validation_failed(tmp_path, monkeypatch):
    """WorkflowValidationFailedError is mapped to error='validation_failed'."""
    from yagra.adapters.inbound import mcp_server
    from yagra.application.use_cases.workflow_persistence import WorkflowValidationFailedError
    from yagra.application.use_cases.workflow_validation_reporter import WorkflowValidationReport

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    def raise_validation(*args: object, **kwargs: object) -> None:
        raise WorkflowValidationFailedError(WorkflowValidationReport(issues=[]))

    monkeypatch.setattr(mcp_server, "save_workflow_with_backup", raise_validation)

    result = mcp_server._tool_apply_update(
        workflow_path=str(wf),
        candidate_yaml=_UPDATED_WORKFLOW_YAML_PROPOSE,
    )

    assert result.get("error") == "validation_failed"
    assert "issues" in result


# ---------------------------------------------------------------------------
# _tool_rollback_update
# ---------------------------------------------------------------------------


def test_tool_rollback_update_success(tmp_path, monkeypatch):
    """Normal rollback returns success=True with expected fields."""
    from yagra.adapters.inbound import mcp_server
    from yagra.application.use_cases.workflow_persistence import WorkflowRollbackResult

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    mock_result = WorkflowRollbackResult(
        restored_revision="rev-old",
        backup_id="bk-001",
        safety_backup_id="bk-safety-001",
    )
    monkeypatch.setattr(mcp_server, "rollback_workflow_from_backup", lambda *a, **k: mock_result)

    result = mcp_server._tool_rollback_update(
        workflow_path=str(wf),
        backup_id="bk-001",
    )

    assert result.get("success") is True
    assert result["restored_revision"] == "rev-old"
    assert result["backup_id"] == "bk-001"
    assert result["safety_backup_id"] == "bk-safety-001"


def test_tool_rollback_update_backup_not_found(tmp_path, monkeypatch):
    """WorkflowBackupNotFoundError is mapped to error='backup_not_found'."""
    from yagra.adapters.inbound import mcp_server
    from yagra.application.services.workflow_file_store import WorkflowBackupNotFoundError

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    def raise_not_found(*args: object, **kwargs: object) -> None:
        raise WorkflowBackupNotFoundError("backup not found: bk-999")

    monkeypatch.setattr(mcp_server, "rollback_workflow_from_backup", raise_not_found)

    result = mcp_server._tool_rollback_update(
        workflow_path=str(wf),
        backup_id="bk-999",
    )

    assert result.get("error") == "backup_not_found"
    assert "message" in result


def test_tool_rollback_update_missing_required_field():
    """Empty workflow_path or backup_id returns an error key."""
    from yagra.adapters.inbound.mcp_server import _tool_rollback_update

    # missing backup_id
    result = _tool_rollback_update(workflow_path="/some/path/workflow.yaml", backup_id="")
    assert "error" in result

    # missing workflow_path
    result2 = _tool_rollback_update(workflow_path="", backup_id="bk-001")
    assert "error" in result2


# ---------------------------------------------------------------------------
# _tool_get_template
# ---------------------------------------------------------------------------


def test_tool_get_template_success():
    """Valid template name returns name and files keys."""
    from yagra.adapters.inbound.mcp_server import _tool_get_template, _tool_list_templates

    templates = _tool_list_templates()
    template_names = [t["name"] for t in templates["templates"]]
    if not template_names:
        return  # no templates available to test

    first_name = template_names[0]
    result = _tool_get_template(first_name)
    assert "error" not in result
    assert result["name"] == first_name
    assert "files" in result
    assert isinstance(result["files"], dict)


def test_tool_get_template_not_found():
    """Non-existent template name returns error='template not found' and available list."""
    from yagra.adapters.inbound.mcp_server import _tool_get_template

    result = _tool_get_template("__nonexistent_template_xyz__")
    assert result["error"] == "template not found"
    assert "name" in result
    assert "available" in result
    assert isinstance(result["available"], list)


# ---------------------------------------------------------------------------
# _tool_validate_workflow with workflow_path: OSError path
# ---------------------------------------------------------------------------


def test_tool_validate_workflow_with_workflow_path_os_error(tmp_path):
    """PermissionError on file read is returned as is_valid=False with schema_error."""
    import stat

    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    workflow_file = tmp_path / "workflow.yaml"
    workflow_file.write_text(
        "version: '1'\nstart_at: n\nend_at: [n]\nnodes:\n  - id: n\n    handler: h\nedges: []\n"
    )
    # Remove read permission
    workflow_file.chmod(0o000)
    try:
        result = _tool_validate_workflow(workflow_path=str(workflow_file))
        assert result["is_valid"] is False
        assert any("schema_error" == i["code"] for i in result["issues"])
    finally:
        workflow_file.chmod(stat.S_IRUSR | stat.S_IWUSR)


# ---------------------------------------------------------------------------
# _tool_explain_workflow with workflow_path: OSError path
# ---------------------------------------------------------------------------


def test_tool_explain_workflow_with_workflow_path_os_error(tmp_path):
    """PermissionError on file read returns error key."""
    import stat

    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    workflow_file = tmp_path / "workflow.yaml"
    workflow_file.write_text(
        "version: '1'\nstart_at: n\nend_at: [n]\nnodes:\n  - id: n\n    handler: h\nedges: []\n"
    )
    workflow_file.chmod(0o000)
    try:
        result = _tool_explain_workflow(workflow_path=str(workflow_file))
        assert "error" in result
    finally:
        workflow_file.chmod(stat.S_IRUSR | stat.S_IWUSR)


# ---------------------------------------------------------------------------
# _tool_propose_update: additional error paths
# ---------------------------------------------------------------------------


def test_tool_propose_update_os_error(tmp_path):
    """PermissionError on current file read returns error key."""
    import stat

    from yagra.adapters.inbound.mcp_server import _tool_propose_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)
    wf.chmod(0o000)
    try:
        result = _tool_propose_update(
            workflow_path=str(wf),
            candidate_yaml=_VALID_WORKFLOW_YAML_PROPOSE,
        )
        assert "error" in result
    finally:
        wf.chmod(stat.S_IRUSR | stat.S_IWUSR)


def test_tool_propose_update_non_mapping_yaml(tmp_path):
    """candidate_yaml that is a list returns error='candidate_yaml must be a mapping'."""
    from yagra.adapters.inbound.mcp_server import _tool_propose_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    result = _tool_propose_update(
        workflow_path=str(wf),
        candidate_yaml="- item1\n- item2\n",
    )
    assert "error" in result
    assert result["error"] == "candidate_yaml must be a mapping"


# ---------------------------------------------------------------------------
# _tool_apply_update: additional error paths
# ---------------------------------------------------------------------------


def test_tool_apply_update_non_mapping_yaml(tmp_path):
    """candidate_yaml that is a list returns error='candidate_yaml must be a mapping'."""
    from yagra.adapters.inbound.mcp_server import _tool_apply_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    result = _tool_apply_update(
        workflow_path=str(wf),
        candidate_yaml="- item1\n- item2\n",
    )
    assert "error" in result
    assert result["error"] == "candidate_yaml must be a mapping"


def test_tool_apply_update_file_not_found():
    """Non-existent workflow_path returns error key with 'not found'."""
    from yagra.adapters.inbound.mcp_server import _tool_apply_update

    result = _tool_apply_update(
        workflow_path="/nonexistent/path/workflow.yaml",
        candidate_yaml=_VALID_WORKFLOW_YAML_PROPOSE,
    )
    assert "error" in result
    assert "not found" in result["error"]


def test_tool_apply_update_current_workflow_os_error(tmp_path):
    """OSError reading current workflow (for auto base_revision) returns error key."""
    import stat

    from yagra.adapters.inbound.mcp_server import _tool_apply_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)
    wf.chmod(0o000)
    try:
        # base_revision=None triggers read of current file
        result = _tool_apply_update(
            workflow_path=str(wf),
            candidate_yaml=_UPDATED_WORKFLOW_YAML_PROPOSE,
            base_revision=None,
        )
        assert "error" in result
    finally:
        wf.chmod(stat.S_IRUSR | stat.S_IWUSR)


def test_tool_apply_update_current_workflow_non_mapping(tmp_path):
    """Current workflow that is not a mapping returns error key."""
    from yagra.adapters.inbound.mcp_server import _tool_apply_update

    wf = tmp_path / "workflow.yaml"
    wf.write_text("- item1\n- item2\n")

    # base_revision=None triggers auto-calculation from current file
    result = _tool_apply_update(
        workflow_path=str(wf),
        candidate_yaml=_VALID_WORKFLOW_YAML_PROPOSE,
        base_revision=None,
    )
    assert "error" in result
    assert result["error"] == "current workflow file is not a mapping"


def test_tool_apply_update_generic_exception(tmp_path, monkeypatch):
    """RuntimeError from save_workflow_with_backup is mapped to error='apply_failed'."""
    from unittest.mock import MagicMock

    from yagra.adapters.inbound import mcp_server

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    mock_save = MagicMock(side_effect=RuntimeError("unexpected error"))
    monkeypatch.setattr(mcp_server, "save_workflow_with_backup", mock_save)

    result = mcp_server._tool_apply_update(
        workflow_path=str(wf),
        candidate_yaml=_UPDATED_WORKFLOW_YAML_PROPOSE,
    )
    assert result.get("error") == "apply_failed"
    assert "message" in result


# ---------------------------------------------------------------------------
# _tool_rollback_update: generic exception path
# ---------------------------------------------------------------------------


def test_tool_rollback_update_generic_exception(tmp_path, monkeypatch):
    """RuntimeError from rollback_workflow_from_backup is mapped to error='rollback_failed'."""
    from unittest.mock import MagicMock

    from yagra.adapters.inbound import mcp_server

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    mock_rollback = MagicMock(side_effect=RuntimeError("unexpected rollback error"))
    monkeypatch.setattr(mcp_server, "rollback_workflow_from_backup", mock_rollback)

    result = mcp_server._tool_rollback_update(
        workflow_path=str(wf),
        backup_id="bk-001",
    )
    assert result.get("error") == "rollback_failed"
    assert "message" in result


# ---------------------------------------------------------------------------
# _tool_get_traces / _tool_analyze_traces: generic exception paths
# ---------------------------------------------------------------------------


def test_tool_get_traces_exception(monkeypatch):
    """Exception from load_traces is returned as error key."""
    from unittest.mock import MagicMock

    # Patch load_traces inside the mcp_server module's function scope via the module
    # _tool_get_traces imports load_traces inside the function body, so we patch the module
    import yagra.application.use_cases.trace_aggregator as trace_agg
    from yagra.adapters.inbound import mcp_server

    mock_load = MagicMock(side_effect=Exception("trace load failure"))
    monkeypatch.setattr(trace_agg, "load_traces", mock_load)

    result = mcp_server._tool_get_traces(trace_dir=".yagra/traces")
    assert "error" in result
    assert "trace load failure" in result["error"]


def test_tool_analyze_traces_exception(monkeypatch):
    """Exception from load_traces in analyze_traces is returned as error key."""
    from unittest.mock import MagicMock

    import yagra.application.use_cases.trace_aggregator as trace_agg
    from yagra.adapters.inbound import mcp_server

    mock_load = MagicMock(side_effect=Exception("analyze trace failure"))
    monkeypatch.setattr(trace_agg, "load_traces", mock_load)

    result = mcp_server._tool_analyze_traces(trace_dir=".yagra/traces")
    assert "error" in result
    assert "analyze trace failure" in result["error"]


# ---------------------------------------------------------------------------
# list_tools(): verify all 11 tool names are present
# ---------------------------------------------------------------------------


def test_list_tools_contains_all_tools():
    """All 11 expected tool names are returned by list_tools via create_mcp_server."""
    import asyncio

    expected_tool_names = {
        "validate_workflow",
        "explain_workflow",
        "list_templates",
        "list_handlers",
        "get_template",
        "get_traces",
        "analyze_traces",
        "propose_update",
        "apply_update",
        "rollback_update",
        "run_golden_tests",
    }

    try:
        from yagra.adapters.inbound.mcp_server import create_mcp_server

        server = create_mcp_server()
    except ImportError:
        return

    list_tools_handler = None
    for handler_name, handler_fn in server.request_handlers.items():
        if "ListTools" in str(handler_name):
            list_tools_handler = handler_fn
            break

    if list_tools_handler is None:
        return

    async def run():
        from mcp.types import ListToolsRequest

        request = ListToolsRequest(method="tools/list")
        result = await list_tools_handler(request)
        return result

    try:
        result = asyncio.run(run())
        tools = result.root.tools if hasattr(result, "root") else []
        tool_names = {t.name for t in tools}
        assert expected_tool_names == tool_names, (
            f"Missing tools: {expected_tool_names - tool_names}, "
            f"Extra tools: {tool_names - expected_tool_names}"
        )
    except Exception:
        pass


# ---------------------------------------------------------------------------
# create_mcp_server with mock mcp: cover list_tools / call_tool closures
# ---------------------------------------------------------------------------


def _build_mock_mcp_server(monkeypatch):
    """Helper: register mock mcp modules and reload mcp_server, return (module, server)."""
    import importlib
    import sys
    from unittest.mock import MagicMock

    class _MockServer:
        """Minimal mock of mcp.server.Server that captures decorated functions."""

        def __init__(self, name: str) -> None:
            self.name = name
            self._list_tools_fn = None
            self._call_tool_fn = None

        def list_tools(self):
            def decorator(fn):
                self._list_tools_fn = fn
                return fn

            return decorator

        def call_tool(self):
            def decorator(fn):
                self._call_tool_fn = fn
                return fn

            return decorator

    mock_mcp_pkg = MagicMock()
    mock_mcp_pkg.server.Server = _MockServer

    mock_types = MagicMock()
    mock_types.Tool = MagicMock(side_effect=lambda **kw: kw)
    mock_types.TextContent = MagicMock(side_effect=lambda type, text: {"type": type, "text": text})

    for mod_name in [
        "mcp",
        "mcp.server",
        "mcp.server.stdio",
        "mcp.server.lowlevel",
        "mcp.server.lowlevel.server",
        "mcp.server.models",
        "mcp.types",
    ]:
        monkeypatch.setitem(sys.modules, mod_name, MagicMock())

    monkeypatch.setitem(sys.modules, "mcp.server", mock_mcp_pkg.server)
    monkeypatch.setitem(sys.modules, "mcp.types", mock_types)

    import yagra.adapters.inbound.mcp_server as mcp_module

    importlib.reload(mcp_module)

    # Patch Server class inside the reloaded module's local import
    monkeypatch.setattr(
        "yagra.adapters.inbound.mcp_server.create_mcp_server",
        lambda: _create_with_mock_server(_MockServer, mock_types, mcp_module),
    )

    server = _create_with_mock_server(_MockServer, mock_types, mcp_module)

    # Restore module
    importlib.reload(mcp_module)

    return mcp_module, server


def _create_with_mock_server(MockServer, mock_types, mcp_module):
    """Directly instantiate MockServer and run the server setup logic."""
    import importlib
    import sys
    from unittest.mock import MagicMock

    # Patch the imports inside create_mcp_server by temporarily injecting mocks
    original_modules = {}
    for mod_name in ["mcp", "mcp.server", "mcp.server.stdio", "mcp.types"]:
        original_modules[mod_name] = sys.modules.get(mod_name)

    mock_mcp_server_module = MagicMock()
    mock_mcp_server_module.Server = MockServer
    sys.modules["mcp.server"] = mock_mcp_server_module
    sys.modules["mcp.server.stdio"] = MagicMock()
    sys.modules["mcp.types"] = mock_types
    sys.modules["mcp"] = MagicMock()

    try:
        importlib.reload(mcp_module)
        server = mcp_module.create_mcp_server()
    finally:
        for mod_name, orig in original_modules.items():
            if orig is None:
                sys.modules.pop(mod_name, None)
            else:
                sys.modules[mod_name] = orig
        importlib.reload(mcp_module)

    return server


def test_create_mcp_server_list_tools_with_mock(monkeypatch):
    """list_tools closure returns 11 Tool dicts when mcp is mocked."""
    import asyncio
    import importlib
    import sys
    from concurrent.futures import ThreadPoolExecutor
    from unittest.mock import MagicMock

    class _MockServer:
        def __init__(self, name):
            self.name = name
            self._list_tools_fn = None
            self._call_tool_fn = None

        def list_tools(self):
            def decorator(fn):
                self._list_tools_fn = fn
                return fn

            return decorator

        def call_tool(self):
            def decorator(fn):
                self._call_tool_fn = fn
                return fn

            return decorator

    mock_types = MagicMock()
    captured_tools = []

    def mock_tool(**kw):
        captured_tools.append(kw)
        return kw

    mock_types.Tool = MagicMock(side_effect=mock_tool)
    mock_types.TextContent = MagicMock(side_effect=lambda type, text: {"type": type, "text": text})

    mock_server_module = MagicMock()
    mock_server_module.Server = _MockServer

    original = {}
    for k in ["mcp", "mcp.server", "mcp.server.stdio", "mcp.types"]:
        original[k] = sys.modules.get(k)

    sys.modules["mcp"] = MagicMock()
    sys.modules["mcp.server"] = mock_server_module
    sys.modules["mcp.server.stdio"] = MagicMock()
    sys.modules["mcp.types"] = mock_types

    try:
        import yagra.adapters.inbound.mcp_server as m

        importlib.reload(m)
        server = m.create_mcp_server()

        assert isinstance(server, _MockServer)
        assert server._list_tools_fn is not None
        assert server._call_tool_fn is not None

        with ThreadPoolExecutor(max_workers=1) as ex:
            tools = ex.submit(asyncio.run, server._list_tools_fn()).result()
        assert len(tools) == 11
        tool_names = {t["name"] for t in tools}
        assert "validate_workflow" in tool_names
        assert "get_template" in tool_names
        assert "propose_update" in tool_names
        assert "apply_update" in tool_names
        assert "rollback_update" in tool_names
        assert "get_traces" in tool_names
        assert "analyze_traces" in tool_names
        assert "run_golden_tests" in tool_names
    finally:
        for k, v in original.items():
            if v is None:
                sys.modules.pop(k, None)
            else:
                sys.modules[k] = v
        importlib.reload(m)


def test_create_mcp_server_call_tool_all_branches(tmp_path, monkeypatch):
    """call_tool closure dispatches to correct _tool_* functions for all tool names."""
    import asyncio
    import importlib
    import json
    import sys
    from concurrent.futures import ThreadPoolExecutor
    from unittest.mock import MagicMock

    class _MockServer:
        def __init__(self, name):
            self.name = name
            self._list_tools_fn = None
            self._call_tool_fn = None

        def list_tools(self):
            def decorator(fn):
                self._list_tools_fn = fn
                return fn

            return decorator

        def call_tool(self):
            def decorator(fn):
                self._call_tool_fn = fn
                return fn

            return decorator

    mock_types = MagicMock()
    mock_types.Tool = MagicMock(side_effect=lambda **kw: kw)

    text_content_instances = []

    def mock_text_content(type, text):
        obj = {"type": type, "text": text}
        text_content_instances.append(obj)
        return obj

    mock_types.TextContent = MagicMock(side_effect=mock_text_content)

    mock_server_module = MagicMock()
    mock_server_module.Server = _MockServer

    original = {}
    for k in ["mcp", "mcp.server", "mcp.server.stdio", "mcp.types"]:
        original[k] = sys.modules.get(k)

    sys.modules["mcp"] = MagicMock()
    sys.modules["mcp.server"] = mock_server_module
    sys.modules["mcp.server.stdio"] = MagicMock()
    sys.modules["mcp.types"] = mock_types

    wf = tmp_path / "workflow.yaml"
    wf.write_text(_VALID_WORKFLOW_YAML_PROPOSE)

    valid_yaml = _VALID_WORKFLOW_YAML_PROPOSE

    try:
        import yagra.adapters.inbound.mcp_server as m
        from yagra.application.use_cases.workflow_persistence import WorkflowSaveResult

        importlib.reload(m)
        server = m.create_mcp_server()

        async def _call_async(name, arguments):
            results = await server._call_tool_fn(name, arguments)
            return json.loads(results[0]["text"])

        def call(name, arguments):
            with ThreadPoolExecutor(max_workers=1) as ex:
                return ex.submit(asyncio.run, _call_async(name, arguments)).result()

        # validate_workflow
        r = call("validate_workflow", {"yaml_content": valid_yaml})
        assert "is_valid" in r

        # explain_workflow
        r = call("explain_workflow", {"yaml_content": valid_yaml})
        assert "entry_point" in r or "error" in r

        # validate_workflow with workflow_path
        r = call("validate_workflow", {"workflow_path": str(wf)})
        assert "is_valid" in r

        # explain_workflow with workflow_path
        r = call("explain_workflow", {"workflow_path": str(wf)})
        assert "entry_point" in r or "error" in r

        # list_templates
        r = call("list_templates", {})
        assert "templates" in r

        # list_handlers
        r = call("list_handlers", {})
        assert "handlers" in r

        # get_template - use first available template
        templates_result = call("list_templates", {})
        if templates_result["templates"]:
            first_name = templates_result["templates"][0]["name"]
            r = call("get_template", {"name": first_name})
            assert "files" in r or "error" in r

        # get_traces (empty dir)
        r = call("get_traces", {"trace_dir": str(tmp_path / "traces"), "limit": 5})
        assert "traces" in r or "error" in r

        # analyze_traces (empty dir)
        r = call("analyze_traces", {"trace_dir": str(tmp_path / "traces")})
        assert "error" in r  # no traces

        # propose_update (file not found)
        r = call(
            "propose_update",
            {"workflow_path": "/nonexistent.yaml", "candidate_yaml": valid_yaml},
        )
        assert "error" in r

        # apply_update - mock save_workflow_with_backup
        mock_result = WorkflowSaveResult(saved_revision="rev1", backup_id="bk-1")
        monkeypatch.setattr(m, "save_workflow_with_backup", lambda *a, **k: mock_result)
        r = call("apply_update", {"workflow_path": str(wf), "candidate_yaml": valid_yaml})
        assert r.get("success") is True or "error" in r

        # rollback_update - missing backup_id
        r = call("rollback_update", {"workflow_path": str(wf), "backup_id": ""})
        assert "error" in r

        # unknown tool
        r = call("__unknown__", {})
        assert "error" in r

    finally:
        for k, v in original.items():
            if v is None:
                sys.modules.pop(k, None)
            else:
                sys.modules[k] = v
        importlib.reload(m)


# ---------------------------------------------------------------------------
# run_mcp_server: lines 942-955 — import path + version fallback + server run
# ---------------------------------------------------------------------------


def test_run_mcp_server_calls_server_run(monkeypatch):
    """run_mcp_server should import mcp, resolve version, and invoke server.run()."""
    import asyncio
    import contextlib

    try:
        import mcp  # noqa: F401
    except ImportError:
        pytest.skip("mcp package not installed")

    from yagra.adapters.inbound import mcp_server as mcp_mod

    run_called = {}

    class _FakeStream:
        pass

    @contextlib.asynccontextmanager
    async def _fake_stdio_server():
        yield _FakeStream(), _FakeStream()

    class _FakeServer:
        def get_capabilities(self, **kwargs):
            return {}

        async def run(self, read_stream, write_stream, init_options):
            run_called["called"] = True
            run_called["server_name"] = init_options.server_name

    monkeypatch.setattr(
        "yagra.adapters.inbound.mcp_server.create_mcp_server",
        lambda: _FakeServer(),
    )

    import mcp.server.stdio as _stdio_mod

    monkeypatch.setattr(_stdio_mod, "stdio_server", _fake_stdio_server)

    asyncio.run(mcp_mod.run_mcp_server())

    assert run_called.get("called") is True
    assert run_called.get("server_name") == "yagra"


def test_run_mcp_server_version_fallback(monkeypatch):
    """run_mcp_server uses '0.0.0' when importlib.metadata raises."""
    import asyncio
    import contextlib

    try:
        import mcp  # noqa: F401
    except ImportError:
        pytest.skip("mcp package not installed")

    from yagra.adapters.inbound import mcp_server as mcp_mod

    version_used = {}

    class _FakeStream:
        pass

    @contextlib.asynccontextmanager
    async def _fake_stdio_server():
        yield _FakeStream(), _FakeStream()

    class _FakeServer:
        def get_capabilities(self, **kwargs):
            return {}

        async def run(self, read_stream, write_stream, init_options):
            version_used["version"] = init_options.server_version

    monkeypatch.setattr(
        "yagra.adapters.inbound.mcp_server.create_mcp_server",
        lambda: _FakeServer(),
    )

    import mcp.server.stdio as _stdio_mod

    monkeypatch.setattr(_stdio_mod, "stdio_server", _fake_stdio_server)

    # Make pkg_version always raise so the fallback "0.0.0" is used
    import importlib.metadata as _meta

    monkeypatch.setattr(_meta, "version", lambda _pkg: (_ for _ in ()).throw(Exception("no pkg")))

    asyncio.run(mcp_mod.run_mcp_server())

    assert version_used.get("version") == "0.0.0"



"""Trace aggregation and summary generation across multiple workflow runs (M-42).

Loads WorkflowRunTrace JSON files from the trace directory, aggregates
per-node and overall statistics, and produces a structured AggregatedSummary.
"""

from __future__ import annotations

import json
import math
import statistics
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any


@dataclass(frozen=True)
class NodeStats:
    """Per-node aggregated statistics across multiple runs.

    Attributes:
        node_id: Unique node identifier from the workflow YAML.
        execution_count: Total number of times this node was executed.
        success_count: Number of executions with status 'success'.
        error_count: Number of executions with status 'error'.
        success_rate: Ratio of successful executions (0.0-1.0).
        avg_duration_ms: Mean execution duration in milliseconds.
        min_duration_ms: Minimum execution duration in milliseconds.
        max_duration_ms: Maximum execution duration in milliseconds.
        p50_duration_ms: Median (50th percentile) execution duration in milliseconds.
        p95_duration_ms: 95th percentile execution duration in milliseconds.
        total_prompt_tokens: Sum of prompt tokens across all LLM calls for this node.
        total_completion_tokens: Sum of completion tokens across all LLM calls for this node.
        total_tokens: Sum of all tokens across all LLM calls for this node.
        avg_tokens_per_call: Average total tokens per LLM call. 0.0 if no LLM calls.
        total_estimated_cost_usd: Aggregated estimated cost in USD. None if no pricing data.
        error_types: Distribution of error types as {error_type: count}.
    """

    node_id: str
    execution_count: int
    success_count: int
    error_count: int
    success_rate: float
    avg_duration_ms: float
    min_duration_ms: float
    max_duration_ms: float
    p50_duration_ms: float
    p95_duration_ms: float
    total_prompt_tokens: int
    total_completion_tokens: int
    total_tokens: int
    avg_tokens_per_call: float
    total_estimated_cost_usd: float | None
    error_types: dict[str, int] = field(default_factory=dict)


@dataclass(frozen=True)
class AggregatedSummary:
    """Summary across multiple workflow runs.

    Attributes:
        workflow_name: Name of the workflow being summarized.
        total_runs: Total number of runs included in this aggregation.
        successful_runs: Number of runs with status 'success'.
        failed_runs: Number of runs with status 'error'.
        run_success_rate: Ratio of successful runs (0.0-1.0).
        avg_total_duration_ms: Mean total run duration in milliseconds.
        min_total_duration_ms: Minimum total run duration in milliseconds.
        max_total_duration_ms: Maximum total run duration in milliseconds.
        total_prompt_tokens: Sum of prompt tokens across all runs.
        total_completion_tokens: Sum of completion tokens across all runs.
        total_tokens: Sum of all tokens across all runs.
        total_estimated_cost_usd: Aggregated estimated cost in USD. None if no pricing data.
        node_stats: Per-node aggregated statistics.
        time_range_start: ISO-8601 timestamp of the earliest run.
        time_range_end: ISO-8601 timestamp of the latest run.
        schema_version: Schema version for this aggregation format.
    """

    workflow_name: str
    total_runs: int
    successful_runs: int
    failed_runs: int
    run_success_rate: float
    avg_total_duration_ms: float
    min_total_duration_ms: float
    max_total_duration_ms: float
    total_prompt_tokens: int
    total_completion_tokens: int
    total_tokens: int
    total_estimated_cost_usd: float | None
    node_stats: list[NodeStats]
    time_range_start: str
    time_range_end: str
    schema_version: str = "1.0"

    def to_dict(self) -> dict[str, Any]:
        """Convert to a JSON-serializable dictionary.

        Returns:
            Dictionary representation with all fields. None values are preserved
            for explicit null representation in JSON output.
        """
        result = asdict(self)
        # asdict handles nested dataclasses automatically.
        # Ensure float values are JSON-safe (no NaN/Inf).
        sanitized = _sanitize_for_json(result)
        assert isinstance(sanitized, dict)  # guaranteed by asdict returning dict
        return sanitized


def _sanitize_for_json(obj: Any) -> Any:
    """Recursively sanitize values for JSON serialization.

    Replaces NaN and Inf float values with None to ensure valid JSON output.

    Args:
        obj: Value to sanitize.

    Returns:
        JSON-safe value.
    """
    if isinstance(obj, float):
        if math.isnan(obj) or math.isinf(obj):
            return None
        return obj
    if isinstance(obj, dict):
        return {k: _sanitize_for_json(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_sanitize_for_json(v) for v in obj]
    return obj


def load_traces(
    trace_dir: str | Path,
    workflow_name: str | None = None,
    limit: int | None = None,
) -> list[dict[str, Any]]:
    """Load trace JSON files from directory.

    Scans the trace directory for JSON files, parses them, sorts by started_at
    in descending order (most recent first), and optionally limits the result count.

    Args:
        trace_dir: Root trace directory (e.g., .yagra/traces).
        workflow_name: Filter by workflow name subdirectory. If None, load from
            all workflow subdirectories.
        limit: Maximum number of traces to load (most recent first). If None,
            load all available traces.

    Returns:
        List of trace dicts parsed from JSON files, sorted by started_at descending.
    """
    trace_path = Path(trace_dir)
    if not trace_path.exists():
        return []

    if workflow_name is not None:
        pattern = f"{workflow_name}/*.json"
    else:
        pattern = "**/*.json"

    traces: list[dict[str, Any]] = []
    for json_file in trace_path.glob(pattern):
        if not json_file.is_file():
            continue
        try:
            data = json.loads(json_file.read_text(encoding="utf-8"))
            if isinstance(data, dict):
                traces.append(data)
        except (json.JSONDecodeError, OSError):
            # Skip malformed or unreadable files silently
            continue

    # Sort by started_at descending (most recent first)
    traces.sort(key=lambda t: t.get("started_at", ""), reverse=True)

    if limit is not None:
        traces = traces[:limit]

    return traces


def aggregate_traces(traces: list[dict[str, Any]]) -> AggregatedSummary:
    """Aggregate multiple trace records into a summary.

    Computes overall run statistics and per-node statistics including duration
    percentiles, token counts, costs, and error type distributions.

    Args:
        traces: List of trace dicts (as loaded from JSON files). Each dict
            should conform to the WorkflowRunTrace.model_dump(mode="json") format.

    Returns:
        AggregatedSummary with per-node and overall statistics.

    Raises:
        ValueError: If traces list is empty.
    """
    if not traces:
        raise ValueError("Cannot aggregate an empty list of traces.")

    # --- Overall run statistics ---
    total_runs = len(traces)
    successful_runs = sum(1 for t in traces if t.get("status") != "error")
    failed_runs = total_runs - successful_runs
    run_success_rate = successful_runs / total_runs if total_runs > 0 else 0.0

    # Workflow name from the first trace
    workflow_name = traces[0].get("workflow_name", "unknown")

    # Duration statistics from the summary.total_duration_ms field
    run_durations: list[float] = []
    for t in traces:
        summary = t.get("summary", {})
        duration = summary.get("total_duration_ms")
        if duration is not None:
            run_durations.append(float(duration))

    avg_total_duration = statistics.mean(run_durations) if run_durations else 0.0
    min_total_duration = min(run_durations) if run_durations else 0.0
    max_total_duration = max(run_durations) if run_durations else 0.0

    # Token and cost aggregation across all runs
    total_prompt_tokens = 0
    total_completion_tokens = 0
    total_tokens = 0
    total_estimated_cost: float | None = None

    for t in traces:
        summary = t.get("summary", {})
        total_prompt_tokens += summary.get("total_prompt_tokens", 0)
        total_completion_tokens += summary.get("total_completion_tokens", 0)
        total_tokens += summary.get("total_tokens", 0)
        cost = summary.get("total_estimated_cost_usd")
        if cost is not None:
            total_estimated_cost = (total_estimated_cost or 0.0) + cost

    # Time range
    started_ats = [t.get("started_at", "") for t in traces if t.get("started_at")]
    ended_ats = [t.get("ended_at", "") for t in traces if t.get("ended_at")]
    time_range_start = min(started_ats) if started_ats else ""
    time_range_end = max(ended_ats) if ended_ats else ""

    # --- Per-node statistics ---
    node_stats = _compute_node_stats(traces)

    return AggregatedSummary(
        workflow_name=workflow_name,
        total_runs=total_runs,
        successful_runs=successful_runs,
        failed_runs=failed_runs,
        run_success_rate=run_success_rate,
        avg_total_duration_ms=avg_total_duration,
        min_total_duration_ms=min_total_duration,
        max_total_duration_ms=max_total_duration,
        total_prompt_tokens=total_prompt_tokens,
        total_completion_tokens=total_completion_tokens,
        total_tokens=total_tokens,
        total_estimated_cost_usd=total_estimated_cost,
        node_stats=node_stats,
        time_range_start=time_range_start,
        time_range_end=time_range_end,
    )


def _compute_node_stats(traces: list[dict[str, Any]]) -> list[NodeStats]:
    """Compute per-node aggregated statistics from trace data.

    Iterates through all nodes across all runs, grouping by node_id, and
    computes duration percentiles, token sums, and error distributions.

    Args:
        traces: List of trace dicts.

    Returns:
        List of NodeStats, one per unique node_id, ordered by first appearance.
    """
    # Collect raw data per node_id, preserving insertion order
    node_data: dict[str, _NodeAccumulator] = {}

    for trace in traces:
        nodes = trace.get("nodes", [])
        for node in nodes:
            node_id = node.get("node_id", "")
            if node_id not in node_data:
                node_data[node_id] = _NodeAccumulator()
            node_data[node_id].add(node)

    return [acc.to_node_stats(node_id) for node_id, acc in node_data.items()]


class _NodeAccumulator:
    """Internal helper that accumulates raw node data for aggregation.

    Collects duration values, token counts, error types, and status counts
    as nodes are added, then produces a final NodeStats via to_node_stats().
    """

    def __init__(self) -> None:
        """Initializes empty accumulator."""
        self.execution_count: int = 0
        self.success_count: int = 0
        self.error_count: int = 0
        self.durations: list[float] = []
        self.prompt_tokens: int = 0
        self.completion_tokens: int = 0
        self.total_tokens_sum: int = 0
        self.llm_call_count: int = 0
        self.estimated_cost: float | None = None
        self.error_types: dict[str, int] = {}

    def add(self, node: dict[str, Any]) -> None:
        """Add a single node trace record to the accumulator.

        Args:
            node: Node trace dict from a WorkflowRunTrace.
        """
        self.execution_count += 1

        status = node.get("status", "")
        if status == "success":
            self.success_count += 1
        elif status == "error":
            self.error_count += 1

        duration = node.get("duration_ms")
        if duration is not None:
            self.durations.append(float(duration))

        llm_call = node.get("llm_call")
        if llm_call is not None:
            self.llm_call_count += 1
            self.prompt_tokens += llm_call.get("prompt_tokens", 0)
            self.completion_tokens += llm_call.get("completion_tokens", 0)
            self.total_tokens_sum += llm_call.get("total_tokens", 0)
            cost = llm_call.get("estimated_cost_usd")
            if cost is not None:
                self.estimated_cost = (self.estimated_cost or 0.0) + cost

        error = node.get("error")
        if error is not None:
            error_type = error.get("error_type", "Unknown")
            self.error_types[error_type] = self.error_types.get(error_type, 0) + 1

    def to_node_stats(self, node_id: str) -> NodeStats:
        """Compute final NodeStats from accumulated data.

        Args:
            node_id: The node identifier.

        Returns:
            Finalized NodeStats dataclass instance.
        """
        success_rate = (
            self.success_count / self.execution_count if self.execution_count > 0 else 0.0
        )

        if self.durations:
            avg_dur = statistics.mean(self.durations)
            min_dur = min(self.durations)
            max_dur = max(self.durations)
            p50_dur = statistics.median(self.durations)
            p95_dur = _percentile_95(self.durations)
        else:
            avg_dur = 0.0
            min_dur = 0.0
            max_dur = 0.0
            p50_dur = 0.0
            p95_dur = 0.0

        avg_tokens = self.total_tokens_sum / self.llm_call_count if self.llm_call_count > 0 else 0.0

        return NodeStats(
            node_id=node_id,
            execution_count=self.execution_count,
            success_count=self.success_count,
            error_count=self.error_count,
            success_rate=success_rate,
            avg_duration_ms=avg_dur,
            min_duration_ms=min_dur,
            max_duration_ms=max_dur,
            p50_duration_ms=p50_dur,
            p95_duration_ms=p95_dur,
            total_prompt_tokens=self.prompt_tokens,
            total_completion_tokens=self.completion_tokens,
            total_tokens=self.total_tokens_sum,
            avg_tokens_per_call=avg_tokens,
            total_estimated_cost_usd=self.estimated_cost,
            error_types=dict(self.error_types),
        )


def _percentile_95(values: list[float]) -> float:
    """Compute the 95th percentile of a list of values.

    Uses statistics.quantiles with n=20 to get the 95th percentile (19th of 20
    quantile boundaries). For lists with fewer than 2 elements, returns the
    single value or 0.0.

    Args:
        values: Non-empty list of numeric values.

    Returns:
        95th percentile value.
    """
    if len(values) == 0:
        return 0.0
    if len(values) == 1:
        return values[0]
    # quantiles(data, n=20) returns 19 cut points; the last one is the 95th percentile
    quantile_points = statistics.quantiles(values, n=20)
    return quantile_points[-1]



"""LLM handler implementation using litellm.

This module provides a generic LLM invocation handler using litellm.
It supports over 100 LLM providers through a unified API.
"""

from typing import Any

import litellm

from yagra.ports.outbound.node_registry import NodeHandler


class LLMHandlerError(Exception):
    """Base exception class for LLM handler execution errors."""

    pass


class LLMHandlerConfigError(LLMHandlerError):
    """Configuration error for LLM handler."""

    pass


class LLMHandlerCallError(LLMHandlerError):
    """Error during LLM invocation."""

    pass


def create_llm_handler(
    retry: int = 3,
    timeout: int = 30,
) -> NodeHandler:
    """Creates an LLM invocation handler.

    Generates a handler that supports over 100 LLM providers using litellm.
    Simply specify `prompt_ref`, `model`, and `output_key` in the YAML
    definition to enable LLM invocation. Variables in the prompt template
    (e.g. ``{query}``) are automatically extracted from state.

    Args:
        retry: Number of retries on API errors (default: 3).
        timeout: Timeout in seconds (default: 30).

    Returns:
        NodeHandler: Handler function that takes (state, params) and returns a dict.

    Examples:
        Basic usage:

        >>> handler = create_llm_handler(retry=3, timeout=30)
        >>> state = {"query": "Hello"}
        >>> params = {
        ...     "prompt": {"system": "You are a helpful assistant", "user": "{query}"},
        ...     "model": {"provider": "openai", "name": "gpt-4", "kwargs": {"temperature": 0.7}},
        ...     "output_key": "response",
        ... }
        >>> result = handler(state, params)
        >>> print(result["response"])  # LLM response

        YAML definition example:

        .. code-block:: yaml

            nodes:
              - id: "chat"
                handler: "llm"
                params:
                  prompt_ref: "prompts/chat.yaml#system"
                  model:
                    provider: "openai"
                    name: "gpt-4"
                    kwargs:
                      temperature: 0.7
                  output_key: "response"
    """

    def handler(state: dict[str, Any], params: dict[str, Any]) -> dict[str, Any]:
        """Invokes the LLM and returns the response.

        Args:
            state: Workflow state dictionary.
            params: Node parameters (prompt, model, output_key).

        Returns:
            dict: Response in the format {output_key: response_text}.

        Raises:
            LLMHandlerConfigError: If required parameters are missing.
            LLMHandlerCallError: If LLM invocation fails.
        """
        from yagra.handlers._llm_common import (
            build_messages,
            extract_llm_params,
            interpolate_prompt,
            llm_retry_loop,
            report_token_usage,
        )

        p = extract_llm_params(params, default_retry=retry)
        system_prompt, user_prompt = interpolate_prompt(
            p.system_prompt_template,
            p.user_prompt_template,
            state,
        )
        messages = build_messages(system_prompt, user_prompt)

        def _call() -> dict[str, Any]:
            response = litellm.completion(
                model=p.litellm_model,
                messages=messages,
                timeout=timeout,
                **p.model_kwargs,
            )

            if not response.choices:
                msg = "LLM returned empty response"
                raise LLMHandlerCallError(msg)

            content = response.choices[0].message.content
            if content is None:
                msg = "LLM returned None content"
                raise LLMHandlerCallError(msg)

            if response.usage is not None:
                report_token_usage(response.usage, p.litellm_model, p.provider)

            return {p.output_key: content}

        return llm_retry_loop(_call, p.effective_retry)

    return handler


LLM_HANDLER_PARAMS_SCHEMA: dict = {
    "type": "object",
    "description": "Parameters for the LLM text output handler created by create_llm_handler",
    "properties": {
        "prompt": {
            "oneOf": [
                {
                    "type": "string",
                    "description": "Prompt text. State values can be expanded using {variable_name}",
                },
                {"type": "object", "description": "Prompt dictionary in role/content format"},
                {"type": "array", "description": "List of multiple messages"},
            ],
            "description": "Prompt definition. Mutually exclusive with prompt_ref",
        },
        "prompt_ref": {
            "type": "string",
            "description": (
                "Path to the prompt file (relative to the workflow YAML). "
                "Use '#key' to select a section from a multi-prompt YAML "
                "(e.g. 'prompts/all.yaml#greet'). "
                "Nested keys use dot notation (e.g. 'prompts/all.yaml#chat.default'). "
                "Mutually exclusive with prompt"
            ),
            "examples": [
                "prompts/translate.yaml#default",
                "./prompts/summarize.md",
                "prompts/multi.yaml#chat.system",
            ],
        },
        "model": {
            "type": "object",
            "description": "LLM model configuration. provider (litellm provider name) and name (model name) are required. Additional litellm parameters can be passed via kwargs",
            "properties": {
                "provider": {
                    "type": "string",
                    "description": "litellm provider name",
                    "examples": ["openai", "anthropic", "google"],
                },
                "name": {
                    "type": "string",
                    "description": "Model name",
                    "examples": ["gpt-4o-mini", "claude-opus-4-6", "gemini-pro"],
                },
                "kwargs": {
                    "type": "object",
                    "description": "Additional parameters to pass to litellm (e.g. temperature)",
                },
            },
            "required": ["provider", "name"],
        },
        "output_key": {
            "type": "string",
            "description": "State key name to store the LLM output. Defaults to 'output'",
            "default": "output",
            "examples": ["translation", "summary", "result"],
        },
    },
    "required": ["model"],
    "oneOf": [
        {"required": ["prompt"]},
        {"required": ["prompt_ref"]},
    ],
}



"""Use case for managing golden test cases (G-20, M-50).

Provides operations to create golden cases from workflow execution traces,
list saved cases, and delete cases. Acts as the orchestration layer between
the domain entities and the persistence port.
"""

from __future__ import annotations

import re
from typing import Any

from yagra.domain.entities.comparison import ComparisonStrategy
from yagra.domain.entities.golden_case import LLM_HANDLER_NAMES, GoldenCase, NodeSnapshot
from yagra.domain.entities.trace import NodeStatus, WorkflowRunTrace
from yagra.ports.outbound.golden_case_repository import GoldenCaseRepositoryPort

_KEBAB_CASE_RE = re.compile(r"^[a-z0-9]+(-[a-z0-9]+)*$")


class GoldenCaseManager:
    """Orchestrates golden case CRUD operations.

    Bridges the gap between raw workflow traces and the golden case
    persistence layer by extracting relevant data from WorkflowRunTrace
    and delegating storage to the repository port.

    Args:
        repository: Persistence backend for golden cases.
    """

    def __init__(self, repository: GoldenCaseRepositoryPort) -> None:
        """Initializes the manager with a repository.

        Args:
            repository: Persistence backend for golden cases.
        """
        self._repository = repository

    def save_from_trace(
        self,
        trace: WorkflowRunTrace,
        case_name: str,
        description: str = "",
        comparison_overrides: dict[str, ComparisonStrategy] | None = None,
    ) -> GoldenCase:
        """Creates and persists a golden case from a successful trace.

        Extracts per-node input/output snapshots, execution path, and
        initial/final state from the trace to build a GoldenCase entity.
        LLM handlers are automatically detected and flagged so that their
        output comparison defaults to structural matching.

        Args:
            trace: A successful WorkflowRunTrace to use as reference.
            case_name: Name for the golden case (kebab-case).
            description: Optional human-readable description.
            comparison_overrides: Per-node comparison strategy overrides.
                Keys are node_ids, values are ComparisonStrategy values.

        Returns:
            The saved GoldenCase.

        Raises:
            ValueError: If trace has failed status or case_name is invalid.
        """
        if trace.status != NodeStatus.SUCCESS:
            msg = f"Cannot create golden case from trace with status '{trace.status}'"
            raise ValueError(msg)

        if not _KEBAB_CASE_RE.match(case_name):
            msg = (
                f"Invalid case_name '{case_name}': must be kebab-case "
                "(lowercase alphanumeric with hyphens, e.g. 'happy-path')"
            )
            raise ValueError(msg)

        overrides = comparison_overrides or {}
        node_snapshots: dict[str, NodeSnapshot] = {}
        execution_path: list[str] = []

        for node_trace in trace.nodes:
            if node_trace.status == NodeStatus.SKIPPED:
                continue
            execution_path.append(node_trace.node_id)

            is_llm = node_trace.handler in LLM_HANDLER_NAMES
            strategy = overrides.get(node_trace.node_id, ComparisonStrategy.AUTO)

            node_snapshots[node_trace.node_id] = NodeSnapshot(
                node_id=node_trace.node_id,
                handler=node_trace.handler,
                is_llm_handler=is_llm,
                input_snapshot=dict(node_trace.input_snapshot),
                output_snapshot=dict(node_trace.output_snapshot),
                comparison_strategy=strategy,
            )

        initial_state = _extract_initial_state(trace)
        final_state = _extract_final_state(trace)

        golden_case = GoldenCase(
            case_name=case_name,
            description=description,
            workflow_name=trace.workflow_name,
            workflow_path=trace.workflow_path or "",
            created_at=trace.started_at,
            source_run_id=str(trace.run_id),
            initial_state=initial_state,
            final_state=final_state,
            execution_path=execution_path,
            node_snapshots=node_snapshots,
        )

        self._repository.save(golden_case)
        return golden_case

    def list_cases(self, workflow_name: str | None = None) -> list[GoldenCase]:
        """Lists saved golden cases, optionally filtered by workflow.

        Args:
            workflow_name: If provided, only returns cases for this workflow.

        Returns:
            List of golden cases sorted by workflow_name then case_name.
        """
        return self._repository.list(workflow_name)

    def delete_case(self, workflow_name: str, case_name: str) -> bool:
        """Deletes a golden case.

        Args:
            workflow_name: Name of the workflow.
            case_name: Name of the case to delete.

        Returns:
            True if the case was deleted, False if it did not exist.
        """
        return self._repository.delete(workflow_name, case_name)


def _extract_initial_state(trace: WorkflowRunTrace) -> dict[str, Any]:
    """Extracts the initial state from the first node's input snapshot.

    Args:
        trace: Workflow run trace.

    Returns:
        Initial state dict. Empty dict if no nodes were executed.
    """
    if not trace.nodes:
        return {}
    return dict(trace.nodes[0].input_snapshot)


def _extract_final_state(trace: WorkflowRunTrace) -> dict[str, Any]:
    """Extracts the final state from the last node's merged state.

    The last node's output is merged with its input to reconstruct
    the complete state at the end of execution.

    Args:
        trace: Workflow run trace.

    Returns:
        Final state dict. Empty dict if no nodes were executed.
    """
    if not trace.nodes:
        return {}
    last = trace.nodes[-1]
    merged = dict(last.input_snapshot)
    merged.update(last.output_snapshot)
    return merged



"""Output port contract for writing execution traces (G-14)."""

from __future__ import annotations

from abc import ABC, abstractmethod

from yagra.domain.entities.trace import WorkflowRunTrace


class TraceSinkError(OSError):
    """Raised when trace persistence fails."""


class TraceSinkPort(ABC):
    """Abstract contract for persisting a WorkflowRunTrace.

    Implementations may write to local files, remote APIs, or in-memory stores.
    The default implementation writes structured JSON to .yagra/traces/.
    """

    @abstractmethod
    def write(self, trace: WorkflowRunTrace) -> str:
        """Persists the trace and returns the destination identifier.

        Args:
            trace: Completed workflow run trace to persist.

        Returns:
            Destination string (file path for local implementations,
            URL for remote implementations, etc.).

        Raises:
            TraceSinkError: If writing fails.
        """



"""Unit tests for TraceCollector (G-14, G-15)."""

from __future__ import annotations

from datetime import UTC, datetime
from typing import Any

import pytest

from yagra.application.use_cases.trace_collector import (
    TraceCollector,
    TraceContext,
    _safe_snapshot,
)
from yagra.domain.entities.trace import NodeStatus


class TestSafeSnapshot:
    def test_primitives_pass_through(self) -> None:
        assert _safe_snapshot("hello") == "hello"
        assert _safe_snapshot(42) == 42
        assert _safe_snapshot(3.14) == pytest.approx(3.14)
        assert _safe_snapshot(True) is True
        assert _safe_snapshot(None) is None

    def test_dict_is_recursed(self) -> None:
        result = _safe_snapshot({"a": 1, "b": "two"})
        assert result == {"a": 1, "b": "two"}

    def test_list_is_recursed(self) -> None:
        result = _safe_snapshot([1, "two", None])
        assert result == [1, "two", None]

    def test_non_serializable_returns_repr(self) -> None:
        class Custom:
            def __repr__(self) -> str:
                return "<Custom instance>"

        result = _safe_snapshot(Custom())
        assert isinstance(result, str)
        assert "Custom" in result

    def test_truncation_applied(self) -> None:
        class LongRepr:
            def __repr__(self) -> str:
                return "x" * 1000

        result = _safe_snapshot(LongRepr(), max_str_len=10)
        assert len(result) == 10


class TestTraceContext:
    def test_current_returns_none_outside_context(self) -> None:
        assert TraceContext.current() is None

    def test_record_and_consume(self) -> None:
        ctx = TraceContext()
        ctx.record_llm_call(
            model="openai/gpt-4o-mini",
            provider="openai",
            prompt_tokens=100,
            completion_tokens=50,
            total_tokens=150,
        )
        result = ctx.consume_llm_call()
        assert result is not None
        assert result.model == "openai/gpt-4o-mini"
        assert result.prompt_tokens == 100
        assert result.total_tokens == 150

    def test_consume_clears_data(self) -> None:
        ctx = TraceContext()
        ctx.record_llm_call(
            model="openai/gpt-4o-mini",
            provider="openai",
            prompt_tokens=10,
            completion_tokens=5,
            total_tokens=15,
        )
        ctx.consume_llm_call()
        assert ctx.consume_llm_call() is None

    def test_consume_without_record_returns_none(self) -> None:
        ctx = TraceContext()
        assert ctx.consume_llm_call() is None

    def test_cost_estimation_applied_for_known_model(self) -> None:
        ctx = TraceContext()
        ctx.record_llm_call(
            model="openai/gpt-4o-mini",
            provider="openai",
            prompt_tokens=1000,
            completion_tokens=500,
            total_tokens=1500,
        )
        result = ctx.consume_llm_call()
        assert result is not None
        assert result.estimated_cost_usd is not None
        assert result.estimated_cost_usd > 0

    def test_cost_estimation_none_for_unknown_model(self) -> None:
        ctx = TraceContext()
        ctx.record_llm_call(
            model="unknown/model-xyz",
            provider="unknown",
            prompt_tokens=100,
            completion_tokens=50,
            total_tokens=150,
        )
        result = ctx.consume_llm_call()
        assert result is not None
        assert result.estimated_cost_usd is None


class TestTraceCollector:
    def _make_collector(self, **kwargs: Any) -> TraceCollector:
        defaults = {
            "workflow_name": "test_workflow",
            "workflow_version": "1.0",
        }
        return TraceCollector(**(defaults | kwargs))

    def test_wrap_node_records_success(self) -> None:
        collector = self._make_collector()

        def simple_handler(state: dict) -> dict:
            return {"output": "hello"}

        wrapped = collector.wrap_node(
            node_id="my_node",
            handler=simple_handler,
            handler_name="custom",
        )
        result = wrapped({"input": "world"})
        assert result == {"output": "hello"}

        now = datetime.now(tz=UTC)
        trace = collector.build_trace(
            started_at=now,
            ended_at=now,
            total_duration_ms=100.0,
        )
        assert len(trace.nodes) == 1
        node = trace.nodes[0]
        assert node.node_id == "my_node"
        assert node.handler == "custom"
        assert node.status == NodeStatus.SUCCESS
        assert node.duration_ms >= 0
        assert node.llm_call is None
        assert node.error is None

    def test_wrap_node_records_error(self) -> None:
        collector = self._make_collector()

        def failing_handler(state: dict) -> dict:
            raise ValueError("test error")

        wrapped = collector.wrap_node(
            node_id="fail_node",
            handler=failing_handler,
            handler_name="custom",
        )
        with pytest.raises(ValueError, match="test error"):
            wrapped({})

        now = datetime.now(tz=UTC)
        trace = collector.build_trace(
            started_at=now,
            ended_at=now,
            total_duration_ms=50.0,
        )
        assert len(trace.nodes) == 1
        node = trace.nodes[0]
        assert node.status == NodeStatus.ERROR
        assert node.error is not None
        assert "test error" in node.error.error_message
        assert "ValueError" in node.error.error_type

    def test_wrap_node_captures_snapshots(self) -> None:
        collector = self._make_collector()

        def handler(state: dict) -> dict:
            return {"result": "done"}

        wrapped = collector.wrap_node("n1", handler, "custom")
        wrapped({"key1": "val1", "key2": 42})

        now = datetime.now(tz=UTC)
        trace = collector.build_trace(started_at=now, ended_at=now, total_duration_ms=10.0)
        node = trace.nodes[0]
        assert node.input_snapshot["key1"] == "val1"
        assert node.input_snapshot["key2"] == 42
        assert "result" in node.output_snapshot

    def test_build_trace_summary_aggregated(self) -> None:
        collector = self._make_collector()

        def handler_a(state: dict) -> dict:
            return {"a": 1}

        def handler_b(state: dict) -> dict:
            return {"b": 2}

        wrapped_a = collector.wrap_node("node_a", handler_a, "custom")
        wrapped_b = collector.wrap_node("node_b", handler_b, "custom")

        wrapped_a({})
        wrapped_b({})

        now = datetime.now(tz=UTC)
        trace = collector.build_trace(started_at=now, ended_at=now, total_duration_ms=200.0)
        assert trace.summary.total_nodes_executed == 2
        assert trace.summary.succeeded_nodes == 2
        assert trace.summary.node_order == ["node_a", "node_b"]

    def test_snapshot_key_filtering(self) -> None:
        collector = self._make_collector(snapshot_keys=frozenset({"allowed"}))

        def handler(state: dict) -> dict:
            return {"allowed": "yes", "secret": "no"}

        wrapped = collector.wrap_node("n1", handler, "custom")
        wrapped({"allowed": "hello", "secret": "hidden"})

        now = datetime.now(tz=UTC)
        trace = collector.build_trace(started_at=now, ended_at=now, total_duration_ms=5.0)
        node = trace.nodes[0]
        assert "allowed" in node.input_snapshot
        assert "secret" not in node.input_snapshot

    def test_include_traceback_false(self) -> None:
        collector = self._make_collector(include_traceback=False)

        def failing_handler(state: dict) -> dict:
            raise RuntimeError("no tb")

        wrapped = collector.wrap_node("n1", failing_handler, "custom")
        with pytest.raises(RuntimeError):
            wrapped({})

        now = datetime.now(tz=UTC)
        trace = collector.build_trace(started_at=now, ended_at=now, total_duration_ms=5.0)
        assert trace.nodes[0].error is not None
        assert trace.nodes[0].error.traceback is None

    def test_overall_status_is_error_when_any_node_fails(self) -> None:
        collector = self._make_collector()

        collector.wrap_node("ok_node", lambda s: {"x": 1}, "custom")({})
        with pytest.raises(ValueError):
            collector.wrap_node(
                "err_node", lambda s: (_ for _ in ()).throw(ValueError("oops")), "custom"
            )({})

        now = datetime.now(tz=UTC)
        trace = collector.build_trace(started_at=now, ended_at=now, total_duration_ms=10.0)
        assert trace.status == NodeStatus.ERROR

    def test_reset_clears_accumulated_node_traces(self) -> None:
        collector = self._make_collector()

        collector.wrap_node("node_a", lambda s: {"ok": True}, "custom")({})
        collector.reset()

        now = datetime.now(tz=UTC)
        trace = collector.build_trace(started_at=now, ended_at=now, total_duration_ms=1.0)
        assert trace.nodes == []



"""Exposes the outbound adapter implementations for Yagra."""

from yagra.adapters.outbound.in_memory_node_registry import InMemoryNodeRegistry
from yagra.adapters.outbound.local_golden_case_store import LocalGoldenCaseStore
from yagra.adapters.outbound.local_trace_sink import LocalTraceSink

__all__ = ["InMemoryNodeRegistry", "LocalGoldenCaseStore", "LocalTraceSink"]



"""E2E integration test for the optimization cycle: Build -> Run & Observe -> Analyze & Propose -> Approve & Update.

This test verifies that the full four-phase optimization cycle works end-to-end
without relying on any external LLM API calls.
"""

from __future__ import annotations

import warnings
from pathlib import Path
from typing import Any

import yaml

from yagra import Yagra
from yagra.adapters.inbound.mcp_server import (
    _tool_analyze_traces,
    _tool_apply_update,
    _tool_get_traces,
    _tool_propose_update,
    _tool_rollback_update,
)
from yagra.application.use_cases.trace_aggregator import aggregate_traces, load_traces

# ---------------------------------------------------------------------------
# Test workflow YAML and helper
# ---------------------------------------------------------------------------

_WORKFLOW_YAML_V1 = """\
version: "1"
start_at: greet
end_at:
  - greet
nodes:
  - id: greet
    handler: greet_handler
edges: []
"""

_WORKFLOW_YAML_V2 = """\
version: "1"
start_at: greet
end_at:
  - farewell
nodes:
  - id: greet
    handler: greet_handler
  - id: farewell
    handler: farewell_handler
edges:
  - source: greet
    target: farewell
"""


def _greet_handler(state: dict[str, Any], params: dict[str, Any]) -> dict[str, Any]:
    """Simple custom handler that returns a greeting message."""
    return {"message": "hello"}


def _farewell_handler(state: dict[str, Any], params: dict[str, Any]) -> dict[str, Any]:
    """Simple custom handler that appends a farewell to state."""
    return {"farewell": "goodbye"}


class TestOptimizationCycleE2E:
    """E2E tests verifying the full optimization cycle.

    Covers the four-phase lifecycle:
      Phase 1: Build  (workflow YAML definition and validation)
      Phase 2: Run & Observe  (execution with trace output)
      Phase 3: Analyze & Propose  (trace analysis and candidate YAML diff)
      Phase 4: Approve & Update  (apply and rollback)
    """

    def test_full_optimization_cycle(self, tmp_path: Path) -> None:
        """Build -> Run & Observe -> Analyze & Propose -> Approve & Update.

        全工程が正常に動作することを検証する。
        """
        # ---------------------------------------------------------------
        # Phase 1: Build - ワークフロー定義と検証
        # ---------------------------------------------------------------

        workflow_path = tmp_path / "workflow.yaml"
        workflow_path.write_text(_WORKFLOW_YAML_V1, encoding="utf-8")

        # MCP validate_workflow でバリデーション
        from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

        validation_result = _tool_validate_workflow(
            yaml_content=_WORKFLOW_YAML_V1,
            base_dir=str(tmp_path),
        )
        assert validation_result["is_valid"] is True, (
            f"Phase 1: workflow validation failed: {validation_result.get('issues')}"
        )

        # ---------------------------------------------------------------
        # Phase 2: Run & Observe - 実行とトレース収集
        # ---------------------------------------------------------------

        trace_dir = tmp_path / ".yagra" / "traces"
        registry = {"greet_handler": _greet_handler}

        yagra = Yagra.from_workflow(
            workflow_path=workflow_path,
            registry=registry,
            observability=True,
        )

        # trace=True でトレースをtmp_path配下に書き出す
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            result = yagra.invoke(
                state={},
                trace=True,
                trace_dir=trace_dir,
            )

        # ワークフロー実行結果の確認
        assert result.get("message") == "hello", f"Phase 2: unexpected workflow result: {result}"

        # .yagra/traces/ 配下にトレースファイルが出力されていることを確認
        trace_files = list(trace_dir.glob("**/*.json"))
        assert len(trace_files) >= 1, (
            f"Phase 2: expected at least 1 trace file in {trace_dir}, found none"
        )

        # ---------------------------------------------------------------
        # Phase 3: Analyze & Propose - トレース分析と変更提案
        # ---------------------------------------------------------------

        # load_traces + aggregate_traces でサマリを生成
        traces = load_traces(trace_dir=str(trace_dir))
        assert len(traces) >= 1, f"Phase 3: load_traces returned no traces from {trace_dir}"

        summary = aggregate_traces(traces)
        assert summary.total_runs >= 1
        assert summary.workflow_name == "workflow"
        assert summary.successful_runs >= 1

        # MCP _tool_get_traces でトレース取得できることを確認
        get_traces_result = _tool_get_traces(
            trace_dir=str(trace_dir),
            workflow_name="workflow",
            limit=20,
        )
        assert "error" not in get_traces_result, (
            f"Phase 3: _tool_get_traces returned error: {get_traces_result.get('error')}"
        )
        assert get_traces_result["count"] >= 1
        assert len(get_traces_result["traces"]) >= 1

        # MCP _tool_analyze_traces でサマリ取得できることを確認
        analyze_result = _tool_analyze_traces(
            trace_dir=str(trace_dir),
            workflow_name="workflow",
        )
        assert "error" not in analyze_result, (
            f"Phase 3: _tool_analyze_traces returned error: {analyze_result.get('error')}"
        )
        assert analyze_result["total_runs"] >= 1
        assert analyze_result["workflow_name"] == "workflow"
        assert len(analyze_result["node_stats"]) >= 1
        # greet ノードの統計が含まれていることを確認
        node_ids = [ns["node_id"] for ns in analyze_result["node_stats"]]
        assert "greet" in node_ids, f"Phase 3: expected 'greet' node in node_stats, got: {node_ids}"

        # MCP _tool_propose_update で候補 YAML の diff とバリデーション結果が返ることを確認
        propose_result = _tool_propose_update(
            workflow_path=str(workflow_path),
            candidate_yaml=_WORKFLOW_YAML_V2,
            reason="add farewell node to complete the greeting cycle",
        )
        assert "error" not in propose_result, (
            f"Phase 3: _tool_propose_update returned error: {propose_result.get('error')}"
        )
        assert propose_result["is_valid"] is True, (
            f"Phase 3: candidate YAML validation failed: {propose_result.get('issues')}"
        )
        assert propose_result["diff"] != "", "Phase 3: expected a non-empty diff between v1 and v2"
        assert propose_result["reason"] == "add farewell node to complete the greeting cycle"

        # ---------------------------------------------------------------
        # Phase 4: Approve & Update - YAML 適用とロールバック
        # ---------------------------------------------------------------

        backup_dir = tmp_path / ".yagra" / "backups"

        # MCP _tool_apply_update で YAML を適用（バックアップ付き）
        apply_result = _tool_apply_update(
            workflow_path=str(workflow_path),
            candidate_yaml=_WORKFLOW_YAML_V2,
            backup_dir=str(backup_dir),
        )
        assert "error" not in apply_result, (
            f"Phase 4: _tool_apply_update returned error: {apply_result.get('error')}"
        )
        assert apply_result.get("success") is True, (
            f"Phase 4: apply_update did not return success=True: {apply_result}"
        )

        # バックアップ ID が返ることを確認
        backup_id = apply_result.get("backup_id")
        assert backup_id is not None and backup_id != "", (
            f"Phase 4: expected a non-empty backup_id, got: {backup_id!r}"
        )

        # YAML が v2 に更新されていることを確認
        applied_content = yaml.safe_load(workflow_path.read_text(encoding="utf-8"))
        assert isinstance(applied_content, dict)
        node_ids_applied = [n["id"] for n in applied_content.get("nodes", [])]
        assert "farewell" in node_ids_applied, (
            f"Phase 4: expected 'farewell' node after apply, got nodes: {node_ids_applied}"
        )

        # MCP _tool_rollback_update でロールバックできることを確認
        rollback_result = _tool_rollback_update(
            workflow_path=str(workflow_path),
            backup_id=backup_id,
            backup_dir=str(backup_dir),
        )
        assert "error" not in rollback_result, (
            f"Phase 4: _tool_rollback_update returned error: {rollback_result.get('error')}"
        )
        assert rollback_result.get("success") is True, (
            f"Phase 4: rollback did not return success=True: {rollback_result}"
        )

        # ロールバック後に元の内容（v1）に戻ることを確認
        rolled_back_content = yaml.safe_load(workflow_path.read_text(encoding="utf-8"))
        assert isinstance(rolled_back_content, dict)
        node_ids_rolled_back = [n["id"] for n in rolled_back_content.get("nodes", [])]
        assert "farewell" not in node_ids_rolled_back, (
            f"Phase 4: 'farewell' node should not be present after rollback, "
            f"got nodes: {node_ids_rolled_back}"
        )
        assert "greet" in node_ids_rolled_back, (
            f"Phase 4: 'greet' node should be present after rollback, "
            f"got nodes: {node_ids_rolled_back}"
        )



"""Unit tests for LocalGoldenCaseStore (G-20)."""

from __future__ import annotations

import json
from datetime import UTC, datetime
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest

from yagra.adapters.outbound.local_golden_case_store import LocalGoldenCaseStore
from yagra.domain.entities.comparison import ComparisonStrategy
from yagra.domain.entities.golden_case import GoldenCase, NodeSnapshot
from yagra.ports.outbound.golden_case_repository import (
    GoldenCaseNotFoundError,
    GoldenCaseRepositoryError,
)


def _make_golden_case(
    case_name: str = "happy-path",
    workflow_name: str = "translate",
    **overrides: Any,
) -> GoldenCase:
    now = datetime.now(tz=UTC)
    defaults: dict[str, Any] = {
        "schema_version": "1.0",
        "case_name": case_name,
        "description": "Test case",
        "workflow_name": workflow_name,
        "workflow_path": f"workflows/{workflow_name}.yaml",
        "created_at": now,
        "source_run_id": "a1b2c3d4-e5f6-7890-abcd-ef1234567890",
        "initial_state": {"text": "hello"},
        "final_state": {"text": "hello", "translated": "こんにちは"},
        "execution_path": ["translate_node"],
        "node_snapshots": {
            "translate_node": NodeSnapshot(
                node_id="translate_node",
                handler="llm",
                is_llm_handler=True,
                input_snapshot={"text": "hello"},
                output_snapshot={"translated": "こんにちは"},
                comparison_strategy=ComparisonStrategy.AUTO,
            ),
        },
        "metadata": {},
    }
    defaults.update(overrides)
    return GoldenCase(**defaults)


class TestLocalGoldenCaseStoreSave:
    def test_save_creates_file(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case()
        path_str = store.save(case)
        assert Path(path_str).exists()

    def test_save_creates_workflow_subdirectory(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case(workflow_name="my_workflow")
        store.save(case)
        assert (tmp_path / "my_workflow").is_dir()

    def test_save_filename_matches_case_name(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case(case_name="my-test-case")
        path_str = store.save(case)
        assert Path(path_str).name == "my-test-case.json"

    def test_save_returns_absolute_path(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case()
        path_str = store.save(case)
        assert Path(path_str).is_absolute()

    def test_saved_json_is_valid(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case()
        path_str = store.save(case)
        content = json.loads(Path(path_str).read_text(encoding="utf-8"))
        assert content["schema_version"] == "1.0"
        assert content["case_name"] == "happy-path"
        assert content["workflow_name"] == "translate"

    def test_save_overwrites_existing(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case1 = _make_golden_case(description="first")
        case2 = _make_golden_case(description="second")
        store.save(case1)
        path_str = store.save(case2)
        content = json.loads(Path(path_str).read_text(encoding="utf-8"))
        assert content["description"] == "second"

    def test_compact_mode(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path, indent=None)
        case = _make_golden_case()
        path_str = store.save(case)
        content = Path(path_str).read_text(encoding="utf-8")
        assert "\n  " not in content

    def test_save_raises_on_mkdir_failure(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case()
        with patch("pathlib.Path.mkdir", side_effect=OSError("Permission denied")):
            with pytest.raises(
                GoldenCaseRepositoryError, match="Failed to create golden case directory"
            ):
                store.save(case)

    def test_save_raises_on_write_failure(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case()
        with patch("pathlib.Path.write_text", side_effect=OSError("Disk full")):
            with pytest.raises(GoldenCaseRepositoryError, match="Failed to write golden case file"):
                store.save(case)


class TestLocalGoldenCaseStoreLoad:
    def test_load_returns_saved_case(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case()
        store.save(case)
        loaded = store.load("translate", "happy-path")
        assert loaded.case_name == "happy-path"
        assert loaded.workflow_name == "translate"

    def test_load_raises_not_found(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        with pytest.raises(GoldenCaseNotFoundError, match="Golden case not found"):
            store.load("translate", "nonexistent")

    def test_load_raises_on_invalid_json(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        workflow_dir = tmp_path / "translate"
        workflow_dir.mkdir(parents=True)
        (workflow_dir / "bad.json").write_text("not valid json", encoding="utf-8")
        with pytest.raises(GoldenCaseRepositoryError, match="Invalid JSON"):
            store.load("translate", "bad")

    def test_load_preserves_node_snapshots(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case()
        store.save(case)
        loaded = store.load("translate", "happy-path")
        assert "translate_node" in loaded.node_snapshots
        snap = loaded.node_snapshots["translate_node"]
        assert snap.handler == "llm"
        assert snap.is_llm_handler is True


class TestLocalGoldenCaseStoreList:
    def test_list_empty_returns_empty(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        assert store.list() == []

    def test_list_nonexistent_base_dir(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path / "nonexistent")
        assert store.list() == []

    def test_list_all_cases(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        store.save(_make_golden_case(case_name="case-a", workflow_name="wf1"))
        store.save(_make_golden_case(case_name="case-b", workflow_name="wf2"))
        cases = store.list()
        assert len(cases) == 2

    def test_list_filtered_by_workflow(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        store.save(_make_golden_case(case_name="case-a", workflow_name="wf1"))
        store.save(_make_golden_case(case_name="case-b", workflow_name="wf2"))
        cases = store.list(workflow_name="wf1")
        assert len(cases) == 1
        assert cases[0].workflow_name == "wf1"

    def test_list_nonexistent_workflow_returns_empty(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        store.save(_make_golden_case())
        assert store.list(workflow_name="nonexistent") == []

    def test_list_sorted_order(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        store.save(_make_golden_case(case_name="case-z", workflow_name="alpha"))
        store.save(_make_golden_case(case_name="case-a", workflow_name="alpha"))
        store.save(_make_golden_case(case_name="case-m", workflow_name="beta"))
        cases = store.list()
        names = [(c.workflow_name, c.case_name) for c in cases]
        assert names == [("alpha", "case-a"), ("alpha", "case-z"), ("beta", "case-m")]

    def test_list_raises_on_corrupt_file(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        workflow_dir = tmp_path / "wf"
        workflow_dir.mkdir(parents=True)
        (workflow_dir / "bad.json").write_text("not json", encoding="utf-8")
        with pytest.raises(GoldenCaseRepositoryError, match="Failed to read"):
            store.list()


class TestLocalGoldenCaseStoreDelete:
    def test_delete_existing_returns_true(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        store.save(_make_golden_case())
        assert store.delete("translate", "happy-path") is True

    def test_delete_removes_file(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        case = _make_golden_case()
        path_str = store.save(case)
        store.delete("translate", "happy-path")
        assert not Path(path_str).exists()

    def test_delete_nonexistent_returns_false(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        assert store.delete("translate", "nonexistent") is False

    def test_delete_raises_on_failure(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        store.save(_make_golden_case())
        with patch("pathlib.Path.unlink", side_effect=OSError("Permission denied")):
            with pytest.raises(GoldenCaseRepositoryError, match="Failed to delete"):
                store.delete("translate", "happy-path")


class TestLocalGoldenCaseStoreExists:
    def test_exists_returns_true(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        store.save(_make_golden_case())
        assert store.exists("translate", "happy-path") is True

    def test_exists_returns_false(self, tmp_path: Path) -> None:
        store = LocalGoldenCaseStore(base_dir=tmp_path)
        assert store.exists("translate", "nonexistent") is False



"""Unit tests for GoldenTestRunner and GoldenCaseManager (G-20, M-50)."""

from __future__ import annotations

from datetime import UTC, datetime
from typing import Any
from uuid import uuid4

import pytest

from yagra.application.use_cases.golden_case_manager import (
    GoldenCaseManager,
    _extract_final_state,
    _extract_initial_state,
)
from yagra.application.use_cases.golden_test_runner import (
    _build_summary,
    _compare_all_nodes,
    _resolve_strategy,
    build_golden_registry,
    compare_node,
    create_mock_llm_handler,
)
from yagra.domain.entities.comparison import ComparisonStrategy
from yagra.domain.entities.golden_case import (
    GoldenCase,
    NodeComparisonResult,
    NodeSnapshot,
)
from yagra.domain.entities.trace import (
    NodeStatus,
    NodeTrace,
    RunSummary,
    WorkflowRunTrace,
)
from yagra.ports.outbound.golden_case_repository import GoldenCaseRepositoryPort

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _make_node_trace(
    node_id: str,
    handler: str,
    input_snapshot: dict[str, Any] | None = None,
    output_snapshot: dict[str, Any] | None = None,
    status: NodeStatus = NodeStatus.SUCCESS,
) -> NodeTrace:
    """Creates a minimal NodeTrace for testing."""
    now = datetime.now(tz=UTC)
    return NodeTrace(
        node_id=node_id,
        handler=handler,
        status=status,
        started_at=now,
        ended_at=now,
        duration_ms=1.0,
        input_snapshot=input_snapshot or {},
        output_snapshot=output_snapshot or {},
    )


def _make_trace(
    nodes: list[NodeTrace],
    workflow_name: str = "test-workflow",
    status: NodeStatus = NodeStatus.SUCCESS,
) -> WorkflowRunTrace:
    """Creates a minimal WorkflowRunTrace for testing."""
    now = datetime.now(tz=UTC)
    return WorkflowRunTrace(
        workflow_name=workflow_name,
        workflow_version="1.0",
        workflow_path="/path/to/workflow.yaml",
        started_at=now,
        ended_at=now,
        status=status,
        nodes=nodes,
        summary=RunSummary(
            total_nodes_executed=len(nodes),
            succeeded_nodes=len([n for n in nodes if n.status == NodeStatus.SUCCESS]),
            failed_nodes=len([n for n in nodes if n.status == NodeStatus.ERROR]),
            total_duration_ms=1.0,
            node_order=[n.node_id for n in nodes],
        ),
    )


def _make_snapshot(
    node_id: str,
    handler: str = "custom",
    is_llm: bool = False,
    input_snap: dict[str, Any] | None = None,
    output_snap: dict[str, Any] | None = None,
    strategy: ComparisonStrategy = ComparisonStrategy.AUTO,
) -> NodeSnapshot:
    """Creates a minimal NodeSnapshot for testing."""
    return NodeSnapshot(
        node_id=node_id,
        handler=handler,
        is_llm_handler=is_llm,
        input_snapshot=input_snap or {},
        output_snapshot=output_snap or {},
        comparison_strategy=strategy,
    )


def _make_golden_case(
    case_name: str = "test-case",
    workflow_name: str = "test-workflow",
    execution_path: list[str] | None = None,
    node_snapshots: dict[str, NodeSnapshot] | None = None,
    initial_state: dict[str, Any] | None = None,
    final_state: dict[str, Any] | None = None,
) -> GoldenCase:
    """Creates a minimal GoldenCase for testing."""
    return GoldenCase(
        case_name=case_name,
        workflow_name=workflow_name,
        workflow_path="workflows/test.yaml",
        created_at=datetime.now(tz=UTC),
        source_run_id=str(uuid4()),
        initial_state=initial_state or {},
        final_state=final_state or {},
        execution_path=execution_path or [],
        node_snapshots=node_snapshots or {},
    )


class FakeRepository(GoldenCaseRepositoryPort):
    """In-memory fake repository for testing."""

    def __init__(self) -> None:
        """Initializes an empty in-memory store."""
        self._store: dict[tuple[str, str], GoldenCase] = {}

    def save(self, case: GoldenCase) -> str:
        self._store[(case.workflow_name, case.case_name)] = case
        return f"/fake/{case.workflow_name}/{case.case_name}.json"

    def load(self, workflow_name: str, case_name: str) -> GoldenCase:
        key = (workflow_name, case_name)
        if key not in self._store:
            msg = f"not found: {workflow_name}/{case_name}"
            raise FileNotFoundError(msg)
        return self._store[key]

    def list(self, workflow_name: str | None = None) -> list[GoldenCase]:
        cases = list(self._store.values())
        if workflow_name is not None:
            cases = [c for c in cases if c.workflow_name == workflow_name]
        return sorted(cases, key=lambda c: (c.workflow_name, c.case_name))

    def delete(self, workflow_name: str, case_name: str) -> bool:
        key = (workflow_name, case_name)
        if key in self._store:
            del self._store[key]
            return True
        return False

    def exists(self, workflow_name: str, case_name: str) -> bool:
        return (workflow_name, case_name) in self._store


# ===========================================================================
# GoldenCaseManager Tests
# ===========================================================================


class TestGoldenCaseManagerSaveFromTrace:
    """Tests for GoldenCaseManager.save_from_trace()."""

    def test_save_from_successful_trace(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        nodes = [
            _make_node_trace(
                "translate",
                "llm",
                input_snapshot={"text": "hello"},
                output_snapshot={"translated": "hola"},
            ),
            _make_node_trace(
                "format",
                "custom_formatter",
                input_snapshot={"translated": "hola"},
                output_snapshot={"formatted": "[hola]"},
            ),
        ]
        trace = _make_trace(nodes)

        result = manager.save_from_trace(trace, "happy-path")

        assert result.case_name == "happy-path"
        assert result.workflow_name == "test-workflow"
        assert len(result.execution_path) == 2
        assert result.execution_path == ["translate", "format"]
        assert "translate" in result.node_snapshots
        assert "format" in result.node_snapshots

        # LLM handler detection
        assert result.node_snapshots["translate"].is_llm_handler is True
        assert result.node_snapshots["format"].is_llm_handler is False

        # Persisted
        assert repo.exists("test-workflow", "happy-path")

    def test_save_extracts_initial_and_final_state(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        nodes = [
            _make_node_trace(
                "step1",
                "custom",
                input_snapshot={"input": "value1"},
                output_snapshot={"result1": "out1"},
            ),
            _make_node_trace(
                "step2",
                "custom",
                input_snapshot={"input": "value1", "result1": "out1"},
                output_snapshot={"result2": "out2"},
            ),
        ]
        trace = _make_trace(nodes)

        result = manager.save_from_trace(trace, "state-test")

        assert result.initial_state == {"input": "value1"}
        assert result.final_state == {
            "input": "value1",
            "result1": "out1",
            "result2": "out2",
        }

    def test_save_rejects_failed_trace(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        trace = _make_trace([], status=NodeStatus.ERROR)

        with pytest.raises(ValueError, match="Cannot create golden case"):
            manager.save_from_trace(trace, "should-fail")

    def test_save_rejects_invalid_case_name(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        trace = _make_trace([])

        with pytest.raises(ValueError, match="kebab-case"):
            manager.save_from_trace(trace, "Invalid Name")

        with pytest.raises(ValueError, match="kebab-case"):
            manager.save_from_trace(trace, "HAS_UNDERSCORE")

        with pytest.raises(ValueError, match="kebab-case"):
            manager.save_from_trace(trace, "")

    def test_save_accepts_valid_kebab_case_names(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        trace = _make_trace(
            [
                _make_node_trace("n1", "custom"),
            ]
        )

        for name in ["simple", "two-words", "a-b-c", "test123", "v2-path"]:
            result = manager.save_from_trace(trace, name)
            assert result.case_name == name

    def test_save_with_comparison_overrides(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        nodes = [
            _make_node_trace("translate", "llm"),
            _make_node_trace("format", "custom"),
        ]
        trace = _make_trace(nodes)

        overrides = {
            "translate": ComparisonStrategy.SKIP,
            "format": ComparisonStrategy.STRUCTURAL,
        }
        result = manager.save_from_trace(trace, "override-test", comparison_overrides=overrides)

        assert result.node_snapshots["translate"].comparison_strategy == ComparisonStrategy.SKIP
        assert result.node_snapshots["format"].comparison_strategy == ComparisonStrategy.STRUCTURAL

    def test_save_with_description(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        trace = _make_trace([_make_node_trace("n", "custom")])
        result = manager.save_from_trace(trace, "desc-test", description="My description")

        assert result.description == "My description"

    def test_save_skips_skipped_nodes(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        nodes = [
            _make_node_trace("executed", "custom"),
            _make_node_trace("skipped", "custom", status=NodeStatus.SKIPPED),
        ]
        trace = _make_trace(nodes)

        result = manager.save_from_trace(trace, "skip-test")

        assert result.execution_path == ["executed"]
        assert "executed" in result.node_snapshots
        assert "skipped" not in result.node_snapshots

    def test_save_detects_structured_llm_handler(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        nodes = [
            _make_node_trace("s1", "structured_llm"),
            _make_node_trace("s2", "streaming_llm"),
        ]
        trace = _make_trace(nodes)

        result = manager.save_from_trace(trace, "llm-types")

        assert result.node_snapshots["s1"].is_llm_handler is True
        assert result.node_snapshots["s2"].is_llm_handler is True


class TestGoldenCaseManagerListAndDelete:
    """Tests for GoldenCaseManager.list_cases() and delete_case()."""

    def test_list_cases_empty(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)
        assert manager.list_cases() == []

    def test_list_cases_returns_all(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        trace = _make_trace([_make_node_trace("n", "custom")])
        manager.save_from_trace(trace, "case-a")
        manager.save_from_trace(trace, "case-b")

        cases = manager.list_cases()
        assert len(cases) == 2

    def test_list_cases_filters_by_workflow(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        trace1 = _make_trace([_make_node_trace("n", "custom")], workflow_name="wf1")
        trace2 = _make_trace([_make_node_trace("n", "custom")], workflow_name="wf2")
        manager.save_from_trace(trace1, "case-a")
        manager.save_from_trace(trace2, "case-b")

        wf1_cases = manager.list_cases(workflow_name="wf1")
        assert len(wf1_cases) == 1
        assert wf1_cases[0].workflow_name == "wf1"

    def test_delete_existing_case(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        trace = _make_trace([_make_node_trace("n", "custom")])
        manager.save_from_trace(trace, "to-delete")

        assert manager.delete_case("test-workflow", "to-delete") is True
        assert manager.list_cases() == []

    def test_delete_nonexistent_case(self) -> None:
        repo = FakeRepository()
        manager = GoldenCaseManager(repo)

        assert manager.delete_case("wf", "nope") is False


# ===========================================================================
# Helper Function Tests
# ===========================================================================


class TestExtractState:
    """Tests for _extract_initial_state and _extract_final_state."""

    def test_initial_state_from_first_node(self) -> None:
        nodes = [
            _make_node_trace("n1", "h", input_snapshot={"a": 1}),
            _make_node_trace("n2", "h", input_snapshot={"a": 1, "b": 2}),
        ]
        trace = _make_trace(nodes)
        assert _extract_initial_state(trace) == {"a": 1}

    def test_initial_state_empty_when_no_nodes(self) -> None:
        trace = _make_trace([])
        assert _extract_initial_state(trace) == {}

    def test_final_state_merges_last_node(self) -> None:
        nodes = [
            _make_node_trace(
                "n1",
                "h",
                input_snapshot={"a": 1},
                output_snapshot={"b": 2},
            ),
            _make_node_trace(
                "n2",
                "h",
                input_snapshot={"a": 1, "b": 2},
                output_snapshot={"c": 3},
            ),
        ]
        trace = _make_trace(nodes)
        assert _extract_final_state(trace) == {"a": 1, "b": 2, "c": 3}

    def test_final_state_empty_when_no_nodes(self) -> None:
        trace = _make_trace([])
        assert _extract_final_state(trace) == {}


# ===========================================================================
# Mock Handler Tests
# ===========================================================================


class TestCreateMockLLMHandler:
    """Tests for create_mock_llm_handler()."""

    def test_returns_output_snapshot(self) -> None:
        snapshot = _make_snapshot(
            "translate",
            "llm",
            is_llm=True,
            output_snap={"translated": "hola"},
        )
        handler = create_mock_llm_handler(snapshot)

        result = handler({"text": "hello"}, {"model": "gpt-4"})
        assert result == {"translated": "hola"}

    def test_ignores_input_state(self) -> None:
        snapshot = _make_snapshot(
            "node",
            "llm",
            is_llm=True,
            output_snap={"out": "value"},
        )
        handler = create_mock_llm_handler(snapshot)

        result1 = handler({"x": 1}, {})
        result2 = handler({"y": 2}, {})
        assert result1 == result2 == {"out": "value"}

    def test_returns_independent_dict(self) -> None:
        snapshot = _make_snapshot(
            "node",
            "llm",
            is_llm=True,
            output_snap={"out": "value"},
        )
        handler = create_mock_llm_handler(snapshot)

        r1 = handler({}, {})
        r1["extra"] = "modified"
        r2 = handler({}, {})
        assert "extra" not in r2


# ===========================================================================
# Registry Building Tests
# ===========================================================================


class TestBuildGoldenRegistry:
    """Tests for build_golden_registry()."""

    def test_mocks_llm_handlers(self) -> None:
        from yagra.adapters.outbound import InMemoryNodeRegistry

        snap_llm = _make_snapshot(
            "translate",
            "llm",
            is_llm=True,
            output_snap={"translated": "hola"},
        )
        snap_custom = _make_snapshot(
            "format",
            "my_formatter",
            is_llm=False,
        )

        golden = _make_golden_case(
            node_snapshots={
                "translate": snap_llm,
                "format": snap_custom,
            },
        )

        def real_formatter(state: dict[str, Any], params: dict[str, Any]) -> dict[str, Any]:
            return {"formatted": state.get("translated", "")}

        base_registry = InMemoryNodeRegistry({"my_formatter": real_formatter})

        result_registry = build_golden_registry(golden, base_registry)

        # LLM handler should be mocked (per-node dispatch via resolve_for_node)
        mock_handler = result_registry.resolve_for_node("llm", "translate")
        assert mock_handler({}, {}) == {"translated": "hola"}

        # Non-LLM handler should be the real one
        real_handler = result_registry.resolve("my_formatter")
        assert real_handler({"translated": "test"}, {}) == {"formatted": "test"}

    def test_per_node_dispatch_for_same_handler(self) -> None:
        from yagra.adapters.outbound import InMemoryNodeRegistry

        snap1 = _make_snapshot("n1", "llm", is_llm=True, output_snap={"a": 1})
        snap2 = _make_snapshot("n2", "llm", is_llm=True, output_snap={"b": 2})

        golden = _make_golden_case(
            node_snapshots={"n1": snap1, "n2": snap2},
        )

        base_registry = InMemoryNodeRegistry()
        result = build_golden_registry(golden, base_registry)

        # Each node gets its own mock handler via resolve_for_node
        handler_n1 = result.resolve_for_node("llm", "n1")
        handler_n2 = result.resolve_for_node("llm", "n2")
        assert handler_n1({}, {}) == {"a": 1}
        assert handler_n2({}, {}) == {"b": 2}


# ===========================================================================
# Node Comparison Tests
# ===========================================================================


class TestCompareNode:
    """Tests for compare_node()."""

    def test_exact_match_passes(self) -> None:
        snapshot = _make_snapshot(
            "format",
            "custom",
            input_snap={"x": 1},
            output_snap={"y": 2},
        )
        result = compare_node(snapshot, {"x": 1}, {"y": 2})

        assert result.status == "pass"
        assert result.input_match is True
        assert result.output_match is True
        assert result.input_diff is None
        assert result.output_diff is None

    def test_input_mismatch_fails(self) -> None:
        snapshot = _make_snapshot(
            "format",
            "custom",
            input_snap={"x": 1},
            output_snap={"y": 2},
        )
        result = compare_node(snapshot, {"x": 999}, {"y": 2})

        assert result.status == "fail"
        assert result.input_match is False
        assert result.output_match is True
        assert result.input_diff is not None

    def test_output_mismatch_fails(self) -> None:
        snapshot = _make_snapshot(
            "format",
            "custom",
            input_snap={"x": 1},
            output_snap={"y": 2},
        )
        result = compare_node(snapshot, {"x": 1}, {"y": 999})

        assert result.status == "fail"
        assert result.input_match is True
        assert result.output_match is False
        assert result.output_diff is not None

    def test_skip_strategy(self) -> None:
        snapshot = _make_snapshot(
            "node",
            "custom",
            input_snap={"x": 1},
            output_snap={"y": 2},
            strategy=ComparisonStrategy.SKIP,
        )
        result = compare_node(snapshot, {"x": 999}, {"y": 999})

        assert result.status == "skip"
        assert result.input_match is None
        assert result.output_match is None

    def test_llm_node_skips_output_comparison(self) -> None:
        snapshot = _make_snapshot(
            "translate",
            "llm",
            is_llm=True,
            input_snap={"text": "hello"},
            output_snap={"translated": "hola"},
        )
        # Even with different output, LLM nodes pass on output
        result = compare_node(snapshot, {"text": "hello"}, {"translated": "different"})

        assert result.output_match is True
        assert result.output_diff is None
        assert result.status == "pass"

    def test_llm_node_input_uses_structural(self) -> None:
        snapshot = _make_snapshot(
            "translate",
            "llm",
            is_llm=True,
            input_snap={"text": "hello"},
            output_snap={"translated": "hola"},
        )
        # Same key, same type, different value -> structural match
        result = compare_node(snapshot, {"text": "different"}, {"translated": "x"})

        assert result.input_match is True
        assert result.status == "pass"

    def test_llm_node_input_structural_type_mismatch(self) -> None:
        snapshot = _make_snapshot(
            "translate",
            "llm",
            is_llm=True,
            input_snap={"text": "hello"},
        )
        # Different type -> structural mismatch
        result = compare_node(snapshot, {"text": 42}, {})

        assert result.input_match is False
        assert result.status == "fail"

    def test_message_on_pass(self) -> None:
        snapshot = _make_snapshot("node", "custom")
        result = compare_node(snapshot, {}, {})
        assert "passed" in result.message

    def test_message_on_fail(self) -> None:
        snapshot = _make_snapshot(
            "node",
            "custom",
            input_snap={"x": 1},
        )
        result = compare_node(snapshot, {"x": 999}, {})
        assert "failed" in result.message
        assert "input" in result.message


# ===========================================================================
# Compare All Nodes Tests
# ===========================================================================


class TestCompareAllNodes:
    """Tests for _compare_all_nodes()."""

    def test_all_matching(self) -> None:
        snap_a = _make_snapshot("a", "custom", input_snap={"x": 1}, output_snap={"y": 2})
        snap_b = _make_snapshot("b", "custom", input_snap={"y": 2}, output_snap={"z": 3})

        golden = _make_golden_case(
            execution_path=["a", "b"],
            node_snapshots={"a": snap_a, "b": snap_b},
        )

        results = _compare_all_nodes(
            golden,
            actual_path=["a", "b"],
            actual_snapshots={
                "a": ({"x": 1}, {"y": 2}),
                "b": ({"y": 2}, {"z": 3}),
            },
        )

        assert len(results) == 2
        assert all(r.status == "pass" for r in results)

    def test_missing_node(self) -> None:
        snap_a = _make_snapshot("a", "custom")
        snap_b = _make_snapshot("b", "custom")

        golden = _make_golden_case(
            execution_path=["a", "b"],
            node_snapshots={"a": snap_a, "b": snap_b},
        )

        results = _compare_all_nodes(
            golden,
            actual_path=["a"],
            actual_snapshots={"a": ({}, {})},
        )

        assert len(results) == 2
        assert results[0].status == "pass"
        assert results[1].status == "missing"
        assert "not executed" in results[1].message

    def test_unexpected_node(self) -> None:
        snap_a = _make_snapshot("a", "custom")

        golden = _make_golden_case(
            execution_path=["a"],
            node_snapshots={"a": snap_a},
        )

        results = _compare_all_nodes(
            golden,
            actual_path=["a", "extra"],
            actual_snapshots={
                "a": ({}, {}),
                "extra": ({"x": 1}, {"y": 2}),
            },
        )

        assert len(results) == 2
        assert results[0].status == "pass"
        assert results[1].status == "unexpected"
        assert results[1].node_id == "extra"

    def test_mixed_results(self) -> None:
        snap_a = _make_snapshot("a", "custom", input_snap={"x": 1})
        snap_b = _make_snapshot("b", "custom")

        golden = _make_golden_case(
            execution_path=["a", "b"],
            node_snapshots={"a": snap_a, "b": snap_b},
        )

        results = _compare_all_nodes(
            golden,
            actual_path=["a"],
            actual_snapshots={"a": ({"x": 999}, {})},
        )

        assert results[0].status == "fail"  # a: input mismatch
        assert results[1].status == "missing"  # b: missing


# ===========================================================================
# Strategy Resolution Tests
# ===========================================================================


class TestResolveStrategy:
    """Tests for _resolve_strategy()."""

    def test_exact_stays_exact(self) -> None:
        assert _resolve_strategy(ComparisonStrategy.EXACT, False) == ComparisonStrategy.EXACT

    def test_structural_stays_structural(self) -> None:
        assert (
            _resolve_strategy(ComparisonStrategy.STRUCTURAL, True) == ComparisonStrategy.STRUCTURAL
        )

    def test_auto_resolves_to_exact_for_non_llm(self) -> None:
        assert _resolve_strategy(ComparisonStrategy.AUTO, False) == ComparisonStrategy.EXACT

    def test_auto_resolves_to_structural_for_llm(self) -> None:
        assert _resolve_strategy(ComparisonStrategy.AUTO, True) == ComparisonStrategy.STRUCTURAL

    def test_skip_stays_skip(self) -> None:
        assert _resolve_strategy(ComparisonStrategy.SKIP, False) == ComparisonStrategy.SKIP


# ===========================================================================
# Summary Building Tests
# ===========================================================================


class TestBuildSummary:
    """Tests for _build_summary()."""

    def test_passed_summary(self) -> None:
        results = [
            NodeComparisonResult(
                node_id="a",
                status="pass",
                strategy_used=ComparisonStrategy.EXACT,
            ),
        ]
        summary = _build_summary("test", True, True, results)
        assert "PASSED" in summary
        assert "1 passed" in summary

    def test_failed_summary(self) -> None:
        results = [
            NodeComparisonResult(
                node_id="a",
                status="fail",
                strategy_used=ComparisonStrategy.EXACT,
            ),
        ]
        summary = _build_summary("test", False, True, results)
        assert "FAILED" in summary
        assert "1 failed" in summary

    def test_path_mismatch_noted(self) -> None:
        summary = _build_summary("test", False, False, [])
        assert "path mismatch" in summary.lower()

    def test_all_statuses_counted(self) -> None:
        results = [
            NodeComparisonResult(
                node_id="a", status="pass", strategy_used=ComparisonStrategy.EXACT
            ),
            NodeComparisonResult(
                node_id="b", status="fail", strategy_used=ComparisonStrategy.EXACT
            ),
            NodeComparisonResult(node_id="c", status="skip", strategy_used=ComparisonStrategy.SKIP),
            NodeComparisonResult(
                node_id="d", status="missing", strategy_used=ComparisonStrategy.EXACT
            ),
            NodeComparisonResult(
                node_id="e",
                status="unexpected",
                strategy_used=ComparisonStrategy.EXACT,
            ),
        ]
        summary = _build_summary("test", False, True, results)
        assert "5 total" in summary
        assert "1 passed" in summary
        assert "1 failed" in summary
        assert "1 skipped" in summary
        assert "1 missing" in summary
        assert "1 unexpected" in summary
